#    Qlattice (https://github.com/jinluchang/qlattice)
#
#    Copyright (C) 2021
#
#    Author: Luchang Jin (ljin.luchang@gmail.com)
#    Author: Masaaki Tomii
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

try:
    from .wick import *
    from . import auto_fac_funcs as aff
except:
    from wick import *
    import auto_fac_funcs as aff

from itertools import permutations

import qlat as q

class Var(Op):

    def __init__(self, name : str):
        Op.__init__(self, "Var")
        self.name = name

    def __repr__(self):
        return f"{self.otype}({self.name!r})"

    def list(self):
        return [self.otype, self.name]

    def __eq__(self, other) -> bool:
        return self.list() == other.list()

### ----

def get_var_name_type(x):
    """
    types include: V_S (wilson matrix), V_G (spin matrix), V_U (color matrix), V_a (c-number)
    """
    if x.startswith("V_S_"):
        return "V_S"
    elif x.startswith("V_U_"):
        return "V_U"
    elif x.startswith("V_tr_"):
        return "V_a"
    elif x.startswith("V_prod_GG_"):
        return "V_G"
    elif x.startswith("V_prod_GS_"):
        return "V_S"
    elif x.startswith("V_prod_SG_"):
        return "V_S"
    elif x.startswith("V_prod_SS_"):
        return "V_S"
    elif x.startswith("V_prod_UU_"):
        return "V_U"
    elif x.startswith("V_prod_UG_"):
        return "V_S"
    elif x.startswith("V_prod_GU_"):
        return "V_S"
    elif x.startswith("V_prod_US_"):
        return "V_S"
    elif x.startswith("V_prod_SU_"):
        return "V_S"
    else:
        assert False

def get_op_type(x):
    """
    ``x`` should be an Op
    return a string which indicate the type of the op
    """
    if not isinstance(x, Op):
        return None
    if x.otype == "S":
        return "S"
    elif x.otype == "G":
        return "G"
    elif x.otype == "U":
        return "U"
    elif x.otype == "Tr":
        return "Tr"
    elif x.otype == "Var":
        return get_var_name_type(x.name)
    else:
        assert False
    return None

def add_positions(s, x):
    """
    ``s`` is set of str
    ``x`` is expr/sub-expr
    """
    if isinstance(x, Term):
        add_positions(s, x.coef)
        for op in x.c_ops:
            add_positions(s, op)
        for op in x.a_ops:
            add_positions(s, op)
    elif isinstance(x, Op):
        if x.otype == "S":
            if isinstance(x.p1, str):
                s.add(x.p1)
            if isinstance(x.p2, str):
                s.add(x.p2)
        elif x.otype == "G":
            if isinstance(x.tag, str):
                s.add(x.tag)
        elif x.otype == "U":
            if isinstance(x.p, str):
                s.add(x.p)
            if isinstance(x.mu, str):
                s.add(x.mu)
        elif x.otype == "Tr":
            for op in x.ops:
                add_positions(s, op)
        elif x.otype == "Qfield":
            if isinstance(x.p, str):
                s.add(x.p)
    elif isinstance(x, Expr):
        for t in x.terms:
            add_positions(s, t)
    elif isinstance(x, ea.Expr):
        for t in x.terms:
            add_positions(s, t)
    elif isinstance(x, ea.Term):
        for f in x.factors:
            add_positions(s, f)
    elif isinstance(x, ea.Factor):
        for v in x.variables:
            s.add(v)

def get_positions(term):
    s = set()
    add_positions(s, term)
    return sorted(list(s))

@q.timer
def collect_position_in_cexpr(named_terms, named_exprs):
    s = set()
    for name, term in named_terms:
        add_positions(s, term)
    for name, expr_list in named_exprs:
        for coef, term_name in expr_list:
            add_positions(s, coef)
    positions = sorted(list(s))
    return positions

@q.timer
def find_common_prod_in_factors(named_exprs):
    """
    return None or (var_name_1, var_name_2,)
    """
    subexpr_count = {}
    for name, expr in named_exprs:
        for ea_coef, term_name in expr:
            assert isinstance(ea_coef, ea.Expr)
            for t in ea_coef.terms:
                x = t.factors
                if len(x) < 1:
                    continue
                for i, f in enumerate(x[:-1]):
                    assert f.otype == "Var"
                    f1 = x[i+1]
                    assert f1.otype == "Var"
                    prod = (f.code, f1.code,)
                    count = subexpr_count.get(prod, 0)
                    subexpr_count[prod] = count + 1
    max_num_repeat = 0
    best_match = None
    for prod, num_repeat in subexpr_count.items():
        if num_repeat > max_num_repeat:
            max_num_repeat = num_repeat
            best_match = prod
    return best_match

@q.timer
def collect_common_prod_in_factors(named_exprs, common_prod, var):
    """
    common_prod = find_common_prod_in_factors(named_exprs)
    var = ea.Factor(name, variables=[], otype="Var")
    """
    for name, expr in named_exprs:
        for ea_coef, term_name in expr:
            assert isinstance(ea_coef, ea.Expr)
            for t in ea_coef.terms:
                x = t.factors
                if len(x) < 1:
                    continue
                for i, f in enumerate(x[:-1]):
                    assert f.otype == "Var"
                    f1 = x[i+1]
                    assert f1.otype == "Var"
                    prod = (f.code, f1.code,)
                    if prod == common_prod:
                        x[i] = var
                        x[i+1] = None
    for name, expr in named_exprs:
        for ea_coef, term_name in expr:
            assert isinstance(ea_coef, ea.Expr)
            for t in ea_coef.terms:
                x = t.factors
                x_new = []
                for v in x:
                    if v is not None:
                        x_new.append(v)
                t.factors = x_new

@q.timer
def find_common_sum_in_factors(named_exprs):
    """
    return None or (var_name_1, var_name_2,)
    """
    subexpr_count = {}
    for name, expr in named_exprs:
        for ea_coef, term_name in expr:
            assert isinstance(ea_coef, ea.Expr)
            x = ea_coef.terms
            if len(x) < 1:
                continue
            for i, t in enumerate(x[:-1]):
                t1 = x[i+1]
                assert t.coef == 1
                assert len(t.factors) == 1
                assert t1.coef == 1
                assert len(t1.factors) == 1
                f = t.factors[0]
                f1 = t1.factors[0]
                assert f.otype == "Var"
                assert f1.otype == "Var"
                pair = (f.code, f1.code,)
                count = subexpr_count.get(pair, 0)
                subexpr_count[pair] = count + 1
    max_num_repeat = 0
    best_match = None
    for pair, num_repeat in subexpr_count.items():
        if num_repeat > max_num_repeat:
            max_num_repeat = num_repeat
            best_match = pair
    return best_match

@q.timer
def collect_common_sum_in_factors(named_exprs, common_pair, var):
    """
    common_pair = find_common_sum_in_factors(named_exprs)
    var = ea.Factor(name, variables=[], otype="Var")
    """
    for name, expr in named_exprs:
        for ea_coef, term_name in expr:
            assert isinstance(ea_coef, ea.Expr)
            x = ea_coef.terms
            if len(x) < 1:
                continue
            for i, t in enumerate(x[:-1]):
                t1 = x[i+1]
                assert t.coef == 1
                assert len(t.factors) == 1
                assert t1.coef == 1
                assert len(t1.factors) == 1
                f = t.factors[0]
                f1 = t1.factors[0]
                assert f.otype == "Var"
                assert f1.otype == "Var"
                pair = (f.code, f1.code,)
                if pair == common_pair:
                    x[i].factors[0] = var
                    x[i+1] = None
    for name, expr in named_exprs:
        for ea_coef, term_name in expr:
            assert isinstance(ea_coef, ea.Expr)
            x = ea_coef.terms
            x_new = []
            for v in x:
                if v is not None:
                    x_new.append(v)
            ea_coef.terms = x_new

@q.timer
def collect_factor_in_cexpr(named_exprs):
    """
    collect the factors in all ea_coef
    collect common sub-expressions in ea_coef
    """
    variables_factor_intermediate = []
    variables_factor = []
    var_nameset = set()
    for name, expr in named_exprs:
        for i, (ea_coef, term_name,) in enumerate(expr):
            expr[i] = (ea.mk_expr(ea.simplified(ea_coef)), term_name,)
    # Add ea.Factor with otype "Expr" to variables_factor_intermediate
    var_counter = 0
    var_dataset = {} # var_dataset[factor_code] = factor_var
    for name, expr in named_exprs:
        for ea_coef, term_name in expr:
            assert isinstance(ea_coef, ea.Expr)
            for t in ea_coef.terms:
                x = t.factors
                for i, f in enumerate(x):
                    if f.otype != "Var":
                        assert f.otype == "Expr"
                        if f.code in var_dataset:
                            x[i] = var_dataset[f.code]
                        else:
                            while True:
                                name = f"V_factor_{var_counter}"
                                var_counter += 1
                                if name not in var_nameset:
                                    break
                            var_nameset.add(name)
                            variables_factor_intermediate.append((name, ea.mk_expr(f),))
                            var = ea.Factor(name, f.variables)
                            x[i] = var
                            var_dataset[f.code] = var
    # Add numerical coef to variables_factor_intermediate
    var_counter = 0
    var_dataset = {} # var_dataset[factor_code] = factor_var
    for name, expr in named_exprs:
        for ea_coef, term_name in expr:
            assert isinstance(ea_coef, ea.Expr)
            for t in ea_coef.terms:
                x = t.coef
                if x == 1:
                    continue
                code = ea.compile_py_complex(x)
                if code in var_dataset:
                    t.coef = 1
                    t.factors.append(var_dataset[code])
                else:
                    while True:
                        name = f"V_factor_coef_{var_counter}"
                        var_counter += 1
                        if name not in var_nameset:
                            break
                    var_nameset.add(name)
                    variables_factor_intermediate.append((name, ea.mk_expr(t.coef),))
                    t.coef = 1
                    var = ea.Factor(name, variables=[], otype="Var")
                    t.factors.append(var)
                    var_dataset[code] = var
    for name, expr in named_exprs:
        for i, (ea_coef, term_name,) in enumerate(expr):
            expr[i] = (ea.mk_expr(ea.simplified(ea_coef)), term_name,)
    # Common product elimination
    var_counter = 0
    var_dataset = {} # var_dataset[(code1, code2,)] = factor_var
    while True:
        prod = find_common_prod_in_factors(named_exprs)
        if prod is None:
            break
        code1, code2 = prod
        var1 = ea.Factor(code1, variables=[], otype="Var")
        var2 = ea.Factor(code2, variables=[], otype="Var")
        prod_expr = ea.mk_expr(var1) * ea.mk_expr(var2)
        while True:
            name = f"V_factor_prod_{var_counter}"
            var_counter += 1
            if name not in var_nameset:
                break
        var_nameset.add(name)
        variables_factor_intermediate.append((name, prod_expr,))
        var = ea.Factor(name, variables=[], otype="Var")
        var_dataset[prod] = var
        collect_common_prod_in_factors(named_exprs, prod, var)
    # Common summation elimination
    var_counter = 0
    var_dataset = {} # var_dataset[(code1, code2,)] = factor_var
    while True:
        pair = find_common_sum_in_factors(named_exprs)
        if pair is None:
            break
        code1, code2 = pair
        var1 = ea.Factor(code1, variables=[], otype="Var")
        var2 = ea.Factor(code2, variables=[], otype="Var")
        pair_expr = ea.mk_expr(var1) + ea.mk_expr(var2)
        while True:
            name = f"V_factor_sum_{var_counter}"
            var_counter += 1
            if name not in var_nameset:
                break
        var_nameset.add(name)
        variables_factor_intermediate.append((name, pair_expr,))
        var = ea.Factor(name, variables=[], otype="Var")
        var_dataset[pair] = var
        collect_common_sum_in_factors(named_exprs, pair, var)
    # Add remaining variables in variables_factor_intermediate to variables_factor
    var_counter = 0
    var_dataset = {} # var_dataset[factor_code] = factor_var
    for name, expr in named_exprs:
        for ea_coef, term_name in expr:
            assert isinstance(ea_coef, ea.Expr)
            for t in ea_coef.terms:
                x = t.factors
                for i, f in enumerate(x):
                    assert f.otype == "Var"
                    if f.code in var_dataset:
                        x[i] = var_dataset[f.code]
                    else:
                        while True:
                            name = f"V_factor_final_{var_counter}"
                            var_counter += 1
                            if name not in var_nameset:
                                break
                        var_nameset.add(name)
                        variables_factor.append((name, ea.mk_expr(f)))
                        var = ea.Factor(name, variables=f.variables, otype=f.otype)
                        x[i] = var
                        var_dataset[f.code] = var
    return variables_factor_intermediate, variables_factor

@q.timer
def collect_prop_in_cexpr(named_terms):
    """
    collect the propagators
    modify the named_terms in-place and return the prop variable definitions as variables
    """
    variables_prop = []
    var_counter = 0
    var_dataset = {} # var_dataset[op_repr] = op_var
    var_nameset = set()
    def add_prop_variables(x):
        nonlocal var_counter
        if isinstance(x, list):
            for i, op in enumerate(x):
                if op.otype in [ "S", ]:
                    op_repr = repr(op)
                    if op_repr in var_dataset:
                        x[i] = var_dataset[op_repr]
                    else:
                        while True:
                            name = f"V_S_{var_counter}"
                            var_counter += 1
                            if name not in var_nameset:
                                break
                        var_nameset.add(name)
                        variables_prop.append((name, op,))
                        var = Var(name)
                        x[i] = var
                        var_dataset[op_repr] = var
                elif op.otype == "Tr":
                    add_prop_variables(op.ops)
        elif isinstance(x, Term):
            add_prop_variables(x.c_ops)
        elif isinstance(x, Expr):
            for t in x.terms:
                add_prop_variables(t)
    for name, term in named_terms:
        add_prop_variables(term)
        term.sort()
    return variables_prop

@q.timer
def collect_color_matrix_in_cexpr(named_terms):
    """
    collect the color matrices
    modify the named_terms in-place and return the color matrix variable definitions as variables
    """
    variables_color_matrix = []
    var_counter = 0
    var_dataset = {} # var_dataset[op_repr] = op_var
    var_nameset = set()
    def add_variables(x):
        nonlocal var_counter
        if isinstance(x, list):
            for i, op in enumerate(x):
                if op.otype in [ "U", ]:
                    op_repr = repr(op)
                    if op_repr in var_dataset:
                        x[i] = var_dataset[op_repr]
                    else:
                        while True:
                            name = f"V_U_{var_counter}"
                            var_counter += 1
                            if name not in var_nameset:
                                break
                        var_nameset.add(name)
                        variables_color_matrix.append((name, op,))
                        var = Var(name)
                        x[i] = var
                        var_dataset[op_repr] = var
                elif op.otype == "Tr":
                    add_variables(op.ops)
        elif isinstance(x, Term):
            add_variables(x.c_ops)
        elif isinstance(x, Expr):
            for t in x.terms:
                add_variables(t)
    for name, term in named_terms:
        add_variables(term)
        term.sort()
    return variables_color_matrix

@q.timer
def collect_tr_in_cexpr(named_terms):
    """
    collect common traces
    modify named_terms in-place and return definitions as variables_tr
    variables_tr = [ (name, value,), ... ]
    possible (name, value,) includes
    ("V_tr_0", op,) where op.otype == "Tr"
    """
    var_nameset = set()
    variables_tr = []
    var_counter = 0
    var_dataset_tr = {} # var_dataset[op_repr] = op_var
    def add_tr_varibles(x):
        nonlocal var_counter
        if isinstance(x, Term):
            add_tr_varibles(x.c_ops)
        elif isinstance(x, Op) and x.otype == "Tr":
            add_tr_varibles(x.ops)
        elif isinstance(x, list):
            for op in x:
                add_tr_varibles(op)
            for i, op in enumerate(x):
                if isinstance(op, Op) and op.otype == "Tr":
                    op_repr = repr(op)
                    if op_repr in var_dataset_tr:
                        x[i] = var_dataset_tr[op_repr]
                    else:
                        while True:
                            name = f"V_tr_{var_counter}"
                            var_counter += 1
                            if name not in var_nameset:
                                break
                        var_nameset.add(name)
                        variables_tr.append((name, op,))
                        var = Var(name)
                        x[i] = var
                        var_dataset_tr[op_repr] = var
    for name, term in named_terms:
        add_tr_varibles(term)
        term.sort()
    return variables_tr

@q.timer
def find_common_subexpr_in_tr(variables_tr):
    """
    return None or (op, op1,)
    """
    subexpr_count = {}
    def add(x, count_added):
        op_repr = repr(x)
        if op_repr in subexpr_count:
            c, op = subexpr_count[op_repr]
            assert x == op
            subexpr_count[op_repr] = (c + count_added, x)
        else:
            subexpr_count[op_repr] = (count_added, x)
    def find_op_pair(x):
        if len(x) < 2:
            return None
        for i, op in enumerate(x):
            factor = 1
            if len(x) > 2:
                factor = 2
            op_type = get_op_type(op)
            if op_type in [ "V_S", "V_G", "V_U", "S", "G", "U", ]:
                op1 = x[(i+1) % len(x)]
                op1_type = get_op_type(op1)
                if op1_type in [ "V_S", "V_G", "V_U", "S", "G", "U", ]:
                    prod = (op, op1,)
                    if op_type in [ "V_G", "G", ] and op1_type in [ "V_G", "G", ]:
                        add(prod, 1.02 * factor)
                    elif op_type in [ "V_U", "U", ] and op1_type in [ "V_U", "U", ]:
                        add(prod, 1.02 * factor)
                    elif op_type in [ "V_U", "U", ] and op1_type in [ "V_G", "G", ]:
                        # do not multiply U with G
                        pass
                    elif op_type in [ "V_G", "G", ] and op1_type in [ "V_U", "U", ]:
                        # do not multiply U with G
                        pass
                    elif op_type in [ "V_G", "G", ] or op1_type in [ "V_G", "G", ]:
                        add(prod, 1.01 * factor)
                    elif op_type in [ "V_U", "U", ] or op1_type in [ "V_U", "U", ]:
                        add(prod, 1.01 * factor)
                    elif op_type in [ "V_S", "S", ] and op1_type in [ "V_S", "S", ]:
                        add(prod, 1 * factor)
                    else:
                        assert False
    def find(x):
        if isinstance(x, list):
            for op in x:
                find(op)
        elif isinstance(x, Op) and x.otype == "Tr" and len(x.ops) >= 2:
            find_op_pair(x.ops)
        elif isinstance(x, Term):
            find(x.c_ops)
        elif isinstance(x, Expr):
            for t in x.terms:
                find(t)
    for name, tr in variables_tr:
        find(tr)
    max_num_repeat = 1.5
    best_match = None
    for num_repeat, op in subexpr_count.values():
        if num_repeat > max_num_repeat:
            max_num_repeat = num_repeat
            best_match = op
    return best_match

@q.timer
def collect_common_subexpr_in_tr(variables_tr, op_common, var):
    op_repr = repr(op_common)
    def replace(x):
        if x is None:
            return None
        elif isinstance(x, list):
            # need to represent the product of the list of operators
            for op in x:
                replace(op)
            if len(x) < 2:
                return None
            for i, op in enumerate(x):
                if isinstance(op, Op) and op.otype in [ "Var", "S", "G", "U", ]:
                    i1 = (i+1) % len(x)
                    op1 = x[i1]
                    if isinstance(op1, Op) and op1.otype in [ "Var", "S", "G", "U", ]:
                        prod = (op, op1,)
                        if repr(prod) == op_repr:
                            x[i1] = None
                            x[i] = var
        elif isinstance(x, Op) and x.otype == "Tr" and len(x.ops) >= 2:
            replace(x.ops)
        elif isinstance(x, Term):
            replace(x.c_ops)
        elif isinstance(x, Expr):
            for t in x.terms:
                replace(t)
    def remove_none(x):
        """
        return a None removed x
        possibly modify in-place
        """
        if isinstance(x, list):
            return [ remove_none(op) for op in x if op is not None ]
        elif isinstance(x, Op):
            if x.otype == "Tr":
                x.ops = remove_none(x.ops)
            return x
        elif isinstance(x, Term):
            x.c_ops = remove_none(x.c_ops)
            x.a_ops = remove_none(x.a_ops)
            return x
        elif isinstance(x, Expr):
            x.terms = [ remove_none(t) for t in x.terms]
            return x
        else:
            assert False
    for name, tr in variables_tr:
        replace(tr)
        remove_none(tr)

@q.timer
def collect_subexpr_in_cexpr(variables_tr):
    """
    collect common sub-expressions
    modify variables_tr in-place and return definitions as variables_prod
    variables_prod = [ (name, value,), ... ]
    possible (name, value,) includes
    ("V_prod_SG_0", [ op, op1, ],) where get_op_type(op) in [ "V_S", "S", ] and get_op_type(op1) in [ "V_G", "G", ]
    """
    var_nameset = set()
    var_counter_dict = {}
    var_counter_dict["V_prod_GG_"] = 0
    var_counter_dict["V_prod_GS_"] = 0
    var_counter_dict["V_prod_SG_"] = 0
    var_counter_dict["V_prod_SS_"] = 0
    var_counter_dict["V_prod_UU_"] = 0
    var_counter_dict["V_prod_UG_"] = 0
    var_counter_dict["V_prod_GU_"] = 0
    var_counter_dict["V_prod_US_"] = 0
    var_counter_dict["V_prod_SU_"] = 0
    variables_prod = []
    while True:
        subexpr = find_common_subexpr_in_tr(variables_tr)
        if subexpr is None:
            break
        op, op1 = subexpr
        op_type = get_op_type(op)
        op1_type = get_op_type(op1)
        assert op_type in [ "V_S", "V_G", "V_U", "S", "G", "U", ]
        assert op1_type in [ "V_S", "V_G", "V_U", "S", "G", "U", ]
        if op_type in [ "V_G", "G", ] and op1_type in [ "V_G", "G", ]:
            name_prefix = "V_prod_GG_"
        elif op_type in [ "V_G", "G", ] and op1_type in [ "V_S", "S", ]:
            name_prefix = "V_prod_GS_"
        elif op_type in [ "V_S", "S", ] and op1_type in [ "V_G", "G", ]:
            name_prefix = "V_prod_SG_"
        elif op_type in [ "V_S", "S", ] and op1_type in [ "V_S", "S", ]:
            name_prefix = "V_prod_SS_"
        elif op_type in [ "V_U", "U", ] and op1_type in [ "V_U", "U", ]:
            name_prefix = "V_prod_UU_"
        elif op_type in [ "V_U", "U", ] and op1_type in [ "V_G", "G", ]:
            name_prefix = "V_prod_UG_"
        elif op_type in [ "V_G", "G", ] and op1_type in [ "V_U", "U", ]:
            name_prefix = "V_prod_GU_"
        elif op_type in [ "V_U", "U", ] and op1_type in [ "V_S", "S", ]:
            name_prefix = "V_prod_US_"
        elif op_type in [ "V_S", "S", ] and op1_type in [ "V_U", "U", ]:
            name_prefix = "V_prod_SU_"
        else:
            assert False
        while True:
            name = f"{name_prefix}{var_counter_dict[name_prefix]}"
            var_counter_dict[name_prefix] += 1
            if name not in var_nameset:
                break
        var_nameset.add(name)
        variables_prod.append((name, subexpr,))
        var = Var(name)
        collect_common_subexpr_in_tr(variables_tr, subexpr, var)
    return variables_prod

class CExpr:

    """
    self.diagram_types
    self.positions
    self.variables_factor_intermediate
    self.variables_factor
    self.variables_prop
    self.variables_color_matrix
    self.variables_prod
    self.variables_tr
    self.named_terms
    self.named_exprs
    #
    self.named_terms[i] = (term_name, Term(c_ops, [], 1),)
    self.named_exprs[i] = (expr_name, [ (ea_coef, term_name,), ... ],)
    self.positions == sorted(list(self.positions))
    """

    def __init__(self):
        self.diagram_types = []
        self.positions = []
        self.variables_factor_intermediate = []
        self.variables_factor = []
        self.variables_prop = []
        self.variables_color_matrix = []
        self.variables_prod = []
        self.variables_tr = []
        self.named_terms = []
        self.named_exprs = []

    @q.timer
    def optimize(self):
        """
        interface function
        """
        self.collect_op()

    @q.timer
    def collect_op(self):
        """
        interface function
        Performing common sub-expression elimination
        Should be called after contract_simplify_compile(*exprs) or mk_cexpr(*exprs)
        The cexpr cannot be evaluated before collect_op!!!
        eval term factor
        """
        for name, term in self.named_terms:
            assert term.coef == 1
            assert term.a_ops == []
        checks = [
                self.variables_factor_intermediate == [],
                self.variables_factor == [],
                self.variables_prop == [],
                self.variables_color_matrix == [],
                self.variables_prod == [],
                self.variables_tr == [],
                ]
        if not all(checks):
            # likely collect_op is already performed
            return
        # collect ea_coef factors into variables
        self.variables_factor_intermediate, self.variables_factor = collect_factor_in_cexpr(self.named_exprs)
        # collect prop expr into variables
        self.variables_prop = collect_prop_in_cexpr(self.named_terms)
        # collect color matrix expr into variables
        self.variables_color_matrix = collect_color_matrix_in_cexpr(self.named_terms)
        # collect trace expr into variables
        self.variables_tr = collect_tr_in_cexpr(self.named_terms)
        # collect common prod into variables
        self.variables_prod = collect_subexpr_in_cexpr(self.variables_tr)

### ----

def increase_type_dict_count(type_dict, key):
    if key in type_dict:
        type_dict[key] += 1
    else:
        type_dict[key] = 1

def drop_tag_last_subscript(tag):
    """
    coordinates with names that are same after dropping last subscript belong to the same permutation group
    """
    return tag.rsplit("_", 1)[0]

def mk_permuting_dicts(pos_list):
    for p_pos_list in permutations(pos_list):
        yield dict(zip(pos_list, p_pos_list))

def get_position_permutation_groups(term):
    """
    allow permutations within the same permutation group
    """
    pos_list = get_positions(term)
    group_dict = dict()
    for pos in pos_list:
        g_pos = drop_tag_last_subscript(pos)
        if g_pos != pos:
            if g_pos not in group_dict:
                group_dict[g_pos] = [ pos, ]
            else:
                group_dict[g_pos].append(pos)
    return group_dict.values()

def mk_combined_permuting_dicts(term):
    """
    produce all possible permutations allowed by permuting within the permutation groups.
    generator of dict, each dict is a map of original coordinate and the permuted coordinate.
    """
    pos_list_list = list(get_position_permutation_groups(term))
    p_dicts_list = [ list(mk_permuting_dicts(pos_list)) for pos_list in pos_list_list ]
    def loop(level):
        if level < 0:
            yield dict()
        else:
            for d1 in loop(level - 1):
                for d2 in p_dicts_list[level]:
                    d = d1.copy()
                    d.update(d2)
                    yield d
    return loop(len(p_dicts_list) - 1)

def loop_term_ops(type_dict, ops):
    for op in ops:
        if op.otype == "S":
            # diagram type elem definition
            increase_type_dict_count(type_dict, (op.p1, op.p2,))
        elif op.otype == "Tr":
            loop_term_ops(type_dict, op.ops)

def get_term_diagram_type_info_no_permutation(term):
    type_dict = dict()
    loop_term_ops(type_dict, term.c_ops)
    return tuple(sorted(type_dict.items()))

def permute_tag(x, p_dict):
    if x in p_dict:
        return p_dict[x]
    else:
        return x

def permute_type_info_entry(x, p_dict):
    ((p1, p2,), n,) = x
    p1 = permute_tag(p1, p_dict)
    p2 = permute_tag(p2, p_dict)
    return ((p1, p2,), n,)

def permute_type_info(type_info, p_dict):
    l = [ permute_type_info_entry(e, p_dict) for e in type_info ]
    return tuple(sorted(l))

def get_term_diagram_type_info(term):
    base_type_info = get_term_diagram_type_info_no_permutation(term)
    min_type_info = base_type_info
    min_type_info_repr = str(min_type_info)
    for p_dict in mk_combined_permuting_dicts(term):
        type_info = permute_type_info(base_type_info, p_dict)
        type_info_repr = repr(type_info)
        if min_type_info is None or type_info_repr < min_type_info_repr:
            min_type_info = type_info
            min_type_info_repr = type_info_repr
    return min_type_info

def filter_diagram_type(expr, diagram_type_dict=None, included_types=None):
    """
    first: drop diagrams with diagram_type_dict[diagram_type] == None
    second:
        if included_types is not None:
            only keep diagrams with diagram_type that diagram_type in included_types.
    ``included_types'' is a list of diagram_type_name
    """
    if diagram_type_dict is None:
        return expr
    new_terms = []
    for term in expr.terms:
        diagram_type = get_term_diagram_type_info(term)
        if diagram_type in diagram_type_dict:
            if diagram_type_dict[diagram_type] is None:
                continue
            if included_types is not None:
                diagram_type_name = diagram_type_dict.get(diagram_type)
                if diagram_type_name not in included_types:
                    continue
        new_terms.append(term)
    included_types_tag = ""
    if included_types is not None:
        included_types_tag = " (" + ','.join(included_types) + ")"
    return Expr(new_terms, expr.description + included_types_tag)

def mk_cexpr(*exprs, diagram_type_dict=None):
    """
    exprs already finished wick contraction,
    otherwise use contract_simplify_compile(*exprs, is_isospin_symmetric_limit, diagram_type_dict)
    !!!if diagram_type_dict[diagram_type] == None: this diagram_type should have already be dropped!!!
    interface function
    """
    if diagram_type_dict is None:
        diagram_type_dict = dict()
    descriptions = [ expr.show() for expr in exprs ]
    # build diagram_types and term names
    diagram_type_counter = 0
    diagram_type_term_dict = dict() # diagram_type_term_dict[repr_term] = diagram_type_name
    term_name_dict = dict() # term_name_dict[term_name] = term
    term_dict = dict() # term_dict[repr(term)] = term_name
    for expr in exprs:
        for term_coef in expr.terms:
            term = Term(term_coef.c_ops, term_coef.a_ops, 1)
            repr_term = repr(term)
            if repr_term in diagram_type_term_dict:
                continue
            diagram_type = get_term_diagram_type_info(term)
            if diagram_type not in diagram_type_dict:
                diagram_type_name = f"ADT{diagram_type_counter}" # ADT is short for "auto diagram type"
                diagram_type_counter += 1
                diagram_type_dict[diagram_type] = diagram_type_name
            diagram_type_name = diagram_type_dict[diagram_type]
            diagram_type_term_dict[repr_term] = diagram_type_name
            assert diagram_type_name is not None
            term_name_counter = 0
            while True:
                term_name = f"term_{diagram_type_name}_{term_name_counter}"
                term_name_counter += 1
                if term_name not in term_name_dict:
                    break
            term_name_dict[term_name] = term
            term_dict[repr_term] = term_name
    # name diagram_types
    diagram_types = []
    for diagram_type, diagram_type_name in diagram_type_dict.items():
        diagram_types.append((diagram_type_name, diagram_type,))
    # name terms
    named_terms = []
    for term_name, term in sorted(term_name_dict.items()):
        named_terms.append((term_name, term,))
    # name exprs
    named_exprs = []
    for i, expr in enumerate(exprs):
        expr_list = []
        typed_expr_list_dict = { name : [] for name, diagram_type in diagram_types }
        for j, term in enumerate(expr.terms):
            coef = term.coef
            term.coef = 1
            repr_term = repr(term)
            diagram_type_name = diagram_type_term_dict[repr_term]
            assert diagram_type_name is not None
            term_name = term_dict[repr_term]
            typed_expr_list_dict[diagram_type_name].append((coef, term_name,))
            expr_list.append((coef, term_name,))
        named_exprs.append((f"{descriptions[i]}  exprs[{i}]", expr_list,))
    # positions
    positions = collect_position_in_cexpr(named_terms, named_exprs)
    # cexpr
    cexpr = CExpr()
    cexpr.diagram_types = diagram_types
    cexpr.positions = positions
    cexpr.named_terms = named_terms
    cexpr.named_exprs = named_exprs
    return cexpr

@q.timer
def contract_simplify(*exprs, is_isospin_symmetric_limit=True, diagram_type_dict=None):
    """
    exprs = [ expr, (expr, *included_types,), ... ]\n
    In case diagram_type_dict is not None, perform the following filter
    If diagram_type_dict[diagram_type] is None: term is removed.
    For (expr, *included_types,), only terms (in expr) with diagram_type that is in included_types is kept
    included_types should be a list/tuple of string
    """
    def func(expr):
        expr = copy.deepcopy(expr)
        if isinstance(expr, tuple):
            expr, *included_types = expr
        elif isinstance(expr, Expr):
            included_types = None
        else:
            assert False
        expr = contract_expr(expr)
        expr.simplify(is_isospin_symmetric_limit = is_isospin_symmetric_limit)
        expr = filter_diagram_type(expr,
                diagram_type_dict=diagram_type_dict,
                included_types=included_types)
        return expr
    return q.parallel_map(func, exprs)

@q.timer
def compile_expr(*exprs, diagram_type_dict = None):
    """
    interface function
    """
    exprs = copy.deepcopy(exprs)
    cexpr = mk_cexpr(*exprs, diagram_type_dict = diagram_type_dict)
    return cexpr

@q.timer
def contract_simplify_compile(*exprs, is_isospin_symmetric_limit = True, diagram_type_dict = None):
    """
    Call ``contract_simplify`` and then ``compile_expr``\n
    This function can be used to construct the first argument of ``cached_comipled_cexpr``.
    e.g. exprs = [ Qb("u", "x", s, c) * Qv("u", "x", s, c) + "u_bar*u", Qb("s", "x", s, c) * Qv("s", "x", s, c) + "s_bar*s", Qb("c", "x", s, c) * Qv("c", "x", s, c) + "c_bar*c", ]
    e.g. exprs = [ mk_pi_p("x2", True) * mk_pi_p("x1") + "(pi   * pi)", mk_j5pi_mu("x2", 3) * mk_pi_p("x1") + "(a_pi * pi)", mk_k_p("x2", True)  * mk_k_p("x1")  + "(k    * k )", mk_j5k_mu("x2", 3)  * mk_k_p("x1")  + "(a_k  * k )", ]
    """
    contracted_simplified_exprs = contract_simplify(
            *exprs,
            is_isospin_symmetric_limit = is_isospin_symmetric_limit,
            diagram_type_dict = diagram_type_dict)
    cexpr = compile_expr(*contracted_simplified_exprs, diagram_type_dict = diagram_type_dict)
    return cexpr

def show_variable_value(value):
    if isinstance(value, list):
        return "*".join(map(show_variable_value, value))
    elif isinstance(value, Var):
        return f"{value.name}"
    elif isinstance(value, G) and value.tag in [ 0, 1, 2, 3, 5, ]:
        tag = { 0: "x", 1: "y", 2: "z", 3: "t", 5: "5", }[value.tag]
        return f"gamma_{tag}"
    elif isinstance(value, G):
        return f"gamma({value.tag})"
    elif isinstance(value, U):
        return f"U({value.tag},{value.p},{value.mu})"
    elif isinstance(value, S):
        return f"S_{value.f}({value.p1},{value.p2})"
    elif isinstance(value, Tr):
        expr = "*".join(map(show_variable_value, value.ops))
        return f"tr({expr})"
    elif isinstance(value, Term):
        if value.coef == 1:
            return "*".join(map(show_variable_value, value.c_ops + value.a_ops))
        else:
            return "*".join(map(show_variable_value, [ f"({value.coef})", ] + value.c_ops + value.a_ops))
    elif isinstance(value, tuple) and len(value) == 2:
        return f"{show_variable_value(value[0])}*{show_variable_value(value[1])}"
    else:
        return f"{value}"

def display_cexpr(cexpr : CExpr):
    """
    return a string
    interface function
    """
    lines = []
    lines.append(f"# Begin CExpr")
    if cexpr.diagram_types:
        lines.append(f"diagram_type_dict = dict()")
        for name, diagram_type in cexpr.diagram_types:
            lines.append(f"diagram_type_dict[{diagram_type}] = {name!r}")
    if cexpr.positions:
        position_vars = ", ".join(cexpr.positions)
        lines.append(f"# Positions:")
        lines.append(f"{position_vars} = {cexpr.positions}")
    if cexpr.variables_prop:
        lines.append(f"# Variables prop:")
    for name, value in cexpr.variables_prop:
        lines.append(f"{name:<30} = {show_variable_value(value)}")
    if cexpr.variables_color_matrix:
        lines.append(f"# Variables color matrix:")
    for name, value in cexpr.variables_color_matrix:
        lines.append(f"{name:<30} = {show_variable_value(value)}")
    if cexpr.variables_factor_intermediate:
        lines.append(f"# Variables factor intermediate:")
    for name, value in cexpr.variables_factor_intermediate:
        lines.append(f"{name:<30} = {show_variable_value(value)}")
    if cexpr.variables_factor:
        lines.append(f"# Variables factor:")
    for name, value in cexpr.variables_factor:
        lines.append(f"{name:<30} = {show_variable_value(value)}")
    if cexpr.variables_prod:
        lines.append(f"# Variables prod:")
    for name, value in cexpr.variables_prod:
        lines.append(f"{name:<30} = {show_variable_value(value)}")
    if cexpr.variables_tr:
        lines.append(f"# Variables tr:")
    for name, value in cexpr.variables_tr:
        lines.append(f"{name:<30} = {show_variable_value(value)}")
    if cexpr.diagram_types:
        lines.append(f"# Diagram type coef:")
    for name, diagram_type in cexpr.diagram_types:
        if name is not None:
            coef_name = f"coef_{name}"
            lines.append(f"{coef_name:<30} = 1")
    if cexpr.named_terms:
        lines.append(f"# Named terms:")
    for idx, (name, term,) in enumerate(cexpr.named_terms):
        name_type = "_".join([ "coef", ] + name.split("_")[1:-1])
        lines.append(f"{name:<30} = {name_type} * {show_variable_value(term)}")
    lines.append(f"terms = [ 0 for i in range({len(cexpr.named_terms)}) ]")
    for idx, (name, term,) in enumerate(cexpr.named_terms):
        lines.append(f"terms[{idx}] = {name}")
    if cexpr.named_exprs:
        lines.append(f"# Named exprs:")
    lines.append(f"exprs = [ 0 for i in range({len(cexpr.named_exprs)}) ]")
    for idx, (name, expr,) in enumerate(cexpr.named_exprs):
        lines.append(f"# {name}")
        for e in expr:
            lines.append(f"exprs[{idx}] += {show_variable_value(e)}")
    lines.append(f"# End CExpr")
    return "\n".join(lines)

@q.timer_verbose
def cexpr_code_gen_py(cexpr : CExpr, *, is_cython = True, is_distillation = False):
    """
    return a string
    interface function
    """
    gen = CExprCodeGenPy(cexpr,
                         is_cython = is_cython,
                         is_distillation = is_distillation)
    return gen.code_gen()

class CExprCodeGenPy:

    """
    self.cexpr
    self.is_cython
    self.is_distillation
    self.var_dict_for_factors
    self.lines
    self.indent
    self.total_sloppy_flops
    #
    flops per complex multiplication: 6
    flops per matrix multiplication: 6 M N L + 2 M L (N-1) ==> 13536 (sc * sc), 4320 (sc * s), 480 (s * s), 3168 (sc * c), 198 (c * c)
    flops per trace 2 (M-1) ==> 22 (sc)
    flops per trace2 6 M N + 2 (M N - 1) ==> 1150 (sc, sc)
    """

    def __init__(self, cexpr, *, is_cython = True, is_distillation = False):
        self.cexpr = cexpr
        self.is_cython = is_cython
        self.is_distillation = is_distillation
        self.var_dict_for_factors = {}
        self.lines = []
        self.indent = 0
        self.total_sloppy_flops = 0
        #
        if self.is_distillation:
            assert self.is_cython == False
        #
        # self.set_var_dict_for_factors();

    def code_gen(self):
        """
        main function
        """
        lines = self.lines
        append = self.append
        append_cy = self.append_cy
        append_py = self.append_py
        if self.is_distillation:
            append(f"from auto_contractor.runtime_distillation import *")
        else:
            append(f"from auto_contractor.runtime import *")
        append_cy(f"import cython")
        append_cy(f"cimport qlat_utils.everything as cc")
        append_cy(f"cimport qlat_utils as qu")
        append_cy(f"cimport libcpp.complex")
        append_cy(f"cimport numpy")
        self.sep()
        self.cexpr_function()
        self.sep()
        self.cexpr_function_get_prop()
        self.sep()
        self.cexpr_function_eval()
        self.sep()
        self.cexpr_function_eval_with_props()
        self.sep()
        self.total_flops()
        return "\n".join(lines)

    def set_var_dict_for_factors(self):
        """
        if this function is called (currently not),
        factor will be referred use array and index,
        instead of its variable name.
        """
        cexpr = self.cexpr
        self.var_dict_for_factors = {}
        var_dict_for_factors = self.var_dict_for_factors
        for idx, (name, value,) in enumerate(cexpr.variables_factor_intermediate):
            assert name.startswith("V_factor_")
            var_dict_for_factors[name] = f"factors_intermediate_view[{idx}]"
        for idx, (name, value,) in enumerate(cexpr.variables_factor):
            assert name.startswith("V_factor_")
            var_dict_for_factors[name] = f"factors_view[{idx}]"

    def gen_expr(self, x):
        """
        return code_str, type_str
        """
        if isinstance(x, (int, float, complex)):
            return f"{x}", "V_a"
        elif isinstance(x, (ea.Expr, ea.Factor)):
            return f"({ea.compile_py(x, self.var_dict_for_factors)})", "V_a"
        assert isinstance(x, Op)
        if x.otype == "S":
            return f"get_prop('{x.f}', {x.p1}, {x.p2})", "V_S"
        elif x.otype == "U":
            return f"get_prop('U', '{x.tag}', {x.p}, {x.mu})", "V_U"
        elif x.otype == "G":
            assert x.s1 == "auto" and x.s2 == "auto"
            assert x.tag in [ 0, 1, 2, 3, 5, ] or isinstance(x.tag, str)
            if self.is_cython:
                if x.tag in [ 0, 1, 2, 3, 5, ]:
                    return f"qu.gamma_matrix_{x.tag}", "V_G"
                else:
                    return f"cc.get_gamma_matrix({x.tag})", "V_G"
            else:
                return f"get_gamma_matrix({x.tag})", "V_G"
        elif x.otype == "Tr":
            if len(x.ops) == 0:
                assert False
            elif len(x.ops) == 1:
                c, t = self.gen_expr(x.ops[0])
                assert t == "V_S"
                self.total_sloppy_flops += 22
                if self.is_cython:
                    return f"cc.matrix_trace({c})", "V_a"
                else:
                    return f"mat_tr_wm({c})", "V_a"
            else:
                c1, t1 = self.gen_expr_prod_list(x.ops[:-1])
                c2, t2 = self.gen_expr(x.ops[-1])
                if t1 == "V_S" and t2 == "V_S":
                    self.total_sloppy_flops += 1150
                    if self.is_cython:
                        return f"cc.matrix_trace({c1}, {c2})", "V_a"
                    else:
                        return f"mat_tr_wm_wm({c1}, {c2})", "V_a"
                elif t1 == "V_S" and t2 == "V_G":
                    if self.is_cython:
                        return f"cc.matrix_trace({c1}, {c2})", "V_a"
                    else:
                        return f"mat_tr_wm_sm({c1}, {c2})", "V_a"
                elif t1 == "V_G" and t2 == "V_S":
                    if self.is_cython:
                        return f"cc.matrix_trace({c1}, {c2})", "V_a"
                    else:
                        return f"mat_tr_sm_wm({c1}, {c2})", "V_a"
                elif t1 == "V_S" and t2 == "V_U":
                    if self.is_cython:
                        return f"cc.matrix_trace({c1}, {c2})", "V_a"
                    else:
                        return f"mat_tr_wm_cm({c1}, {c2})", "V_a"
                elif t1 == "V_U" and t2 == "V_S":
                    if self.is_cython:
                        return f"cc.matrix_trace({c1}, {c2})", "V_a"
                    else:
                        return f"mat_tr_cm_wm({c1}, {c2})", "V_a"
                else:
                    assert False
        elif x.otype == "Var":
            if x.name.startswith("V_S_") or x.name.startswith("V_U_"):
                if self.is_cython:
                    return f"p_{x.name}[0]", get_var_name_type(x.name)
                else:
                    return f"p_{x.name}", get_var_name_type(x.name)
            else:
                return f"{x.name}", get_var_name_type(x.name)

    def gen_expr_prod(self, ct1, ct2):
        c1, t1 = ct1
        c2, t2 = ct2
        if c1 == "(1+0j)":
            assert t1 == "V_a"
            return ct2
        elif c2 == "(1+0j)":
            assert t2 == "V_a"
            return ct1
        elif t1 == "V_S" and t2 == "V_S":
            self.total_sloppy_flops += 13536
            if self.is_cython:
                return f"{c1} * {c2}", "V_S"
            else:
                return f"mat_mul_wm_wm({c1}, {c2})", "V_S"
        #
        elif t1 == "V_S" and t2 == "V_a":
            if self.is_cython:
                return f"{c1} * {c2}", "V_S"
            else:
                return f"mat_mul_a_wm({c2}, {c1})", "V_S"
        elif t1 == "V_a" and t2 == "V_S":
            if self.is_cython:
                return f"{c1} * {c2}", "V_S"
            else:
                return f"mat_mul_a_wm({c1}, {c2})", "V_S"
        elif t1 == "V_a" and t2 == "V_a":
            return f"{c1} * {c2}", "V_a"
        #
        elif t1 == "V_S" and t2 == "V_G":
            self.total_sloppy_flops += 4320
            if self.is_cython:
                return f"{c1} * {c2}", "V_S"
            else:
                return f"mat_mul_wm_sm({c1}, {c2})", "V_S"
        elif t1 == "V_G" and t2 == "V_S":
            self.total_sloppy_flops += 4320
            if self.is_cython:
                return f"{c1} * {c2}", "V_S"
            else:
                return f"mat_mul_sm_wm({c1}, {c2})", "V_S"
        elif t1 == "V_G" and t2 == "V_G":
            self.total_sloppy_flops += 480
            if self.is_cython:
                return f"{c1} * {c2}", "V_G"
            else:
                return f"mat_mul_sm_sm({c1}, {c2})", "V_G"
        elif t1 == "V_G" and t2 == "V_a":
            if self.is_cython:
                return f"{c1} * {c2}", "V_G"
            else:
                return f"mat_mul_a_sm({c2}, {c1})", "V_G"
        elif t1 == "V_a" and t2 == "V_G":
            if self.is_cython:
                return f"{c1} * {c2}", "V_G"
            else:
                return f"mat_mul_a_sm({c1}, {c2})", "V_G"
        #
        elif t1 == "V_S" and t2 == "V_U":
            self.total_sloppy_flops += 3168
            if self.is_cython:
                return f"{c1} * {c2}", "V_S"
            else:
                return f"mat_mul_wm_cm({c1}, {c2})", "V_S"
        elif t1 == "V_U" and t2 == "V_S":
            self.total_sloppy_flops += 3168
            if self.is_cython:
                return f"{c1} * {c2}", "V_S"
            else:
                return f"mat_mul_cm_wm({c1}, {c2})", "V_S"
        elif t1 == "V_U" and t2 == "V_U":
            self.total_sloppy_flops += 198
            if self.is_cython:
                return f"{c1} * {c2}", "V_U"
            else:
                return f"mat_mul_cm_cm({c1}, {c2})", "V_U"
        elif t1 == "V_U" and t2 == "V_a":
            if self.is_cython:
                return f"{c1} * {c2}", "V_U"
            else:
                return f"mat_mul_a_cm({c2}, {c1})", "V_U"
        elif t1 == "V_a" and t2 == "V_U":
            if self.is_cython:
                return f"{c1} * {c2}", "V_U"
            else:
                return f"mat_mul_a_cm({c1}, {c2})", "V_U"
        #
        else:
            print(ct1, ct2)
            assert False

    def gen_expr_prod_list(self, x_list):
        if len(x_list) == 0:
            return f"1", "V_a"
        elif len(x_list) == 1:
            return self.gen_expr(x_list[0])
        else:
            assert len(x_list) > 1
            return self.gen_expr_prod(self.gen_expr_prod_list(x_list[:-1]), self.gen_expr(x_list[-1]))

    def append(self, line):
        lines = self.lines
        if line == "":
            lines.append("")
        else:
            lines.append(self.indent * ' ' + line)

    def append_py(self, line):
        if self.is_cython:
            return
        lines = self.lines
        if line == "":
            lines.append("# Python only")
        else:
            lines.append(self.indent * ' ' + line + " # Python only")

    def append_cy(self, line):
        if not self.is_cython:
            return
        lines = self.lines
        if line == "":
            lines.append("# Cython")
        else:
            lines.append(self.indent * ' ' + line + " # Cython")

    def sep(self):
        append = self.append
        append(f"")
        append(f"### ----")
        append(f"")

    def cexpr_function(self):
        append = self.append
        append_cy = self.append_cy
        append_py = self.append_py
        append(f"@timer")
        append(f"def cexpr_function(*, positions_dict, get_prop, is_ama_and_sloppy=False):")
        self.indent += 4
        append(f"# get_props")
        append(f"props, cms, factors = cexpr_function_get_prop(positions_dict, get_prop)")
        append(f"# eval")
        append(f"ama_val = cexpr_function_eval(positions_dict, props, cms, factors)")
        append(f"# extract sloppy val")
        append(f"val_sloppy = ama_extract(ama_val, is_sloppy = True)")
        append(f"# extract AMA val")
        append(f"val_ama = ama_extract(ama_val)")
        append(f"# return")
        append(f"if is_ama_and_sloppy:")
        append(f"    # return both AMA corrected results and sloppy results")
        append(f"    return val_ama, val_sloppy")
        append(f"else:")
        append(f"    # return AMA corrected results by default")
        append(f"    return val_ama")
        self.indent -= 4

    def cexpr_function_get_prop(self):
        append = self.append
        append_cy = self.append_cy
        append_py = self.append_py
        cexpr = self.cexpr
        append(f"@timer_flops")
        append_cy(f"@cython.boundscheck(False)")
        append_cy(f"@cython.wraparound(False)")
        append(f"def cexpr_function_get_prop(positions_dict, get_prop):")
        self.indent += 4
        append(f"# set positions")
        for position_var in cexpr.positions:
            if position_var in aff.auto_fac_funcs_list:
                append(f"{position_var} = aff.{position_var}")
            else:
                append(f"{position_var} = positions_dict['{position_var}']")
        append(f"# get prop")
        for name, value in cexpr.variables_prop:
            assert name.startswith("V_S_")
            x = value
            assert isinstance(x, Op)
            assert x.otype == "S"
            c, t = self.gen_expr(x)
            assert t == "V_S"
            # potentially be AMA prop (cannot use cdef in cython)
            append(f"{name} = {c}")
        append(f"# get color matrix")
        for name, value in cexpr.variables_color_matrix:
            assert name.startswith("V_U_")
            x = value
            assert isinstance(x, Op)
            assert x.otype == "U"
            c, t = self.gen_expr(x)
            assert t == "V_U"
            append_cy(f"cdef qu.ColorMatrix {name} = {c}")
            append_py(f"{name} = {c}")
        append(f"# set props for return")
        append(f"props = [")
        self.indent += 4
        for name, value in cexpr.variables_prop:
            append(f"{name},")
        append(f"]")
        self.indent -= 4
        append(f"# set color matrix for return")
        append(f"cms = [")
        self.indent += 4
        for name, value in cexpr.variables_color_matrix:
            append(f"{name},")
        append(f"]")
        self.indent -= 4
        append(f"# set intermediate factors")
        for idx, (name, value,) in enumerate(cexpr.variables_factor_intermediate):
            assert name.startswith("V_factor_")
            x = value
            assert isinstance(x, ea.Expr)
            c, t = self.gen_expr(x)
            assert t == "V_a"
            append(f"# {idx} {name}")
            append_cy(f"cdef cc.PyComplex {name} = {c}")
            append_py(f"{name} = {c}")
        append(f"# declare factors")
        append_cy(f"cdef numpy.ndarray[numpy.complex128_t] factors")
        append(f"factors = np.zeros({len(cexpr.variables_factor)}, dtype=np.complex128)")
        append_cy(f"cdef cc.PyComplex[:] factors_view = factors")
        append_py(f"factors_view = factors")
        append(f"# set factors")
        for idx, (name, value,) in enumerate(cexpr.variables_factor):
            assert name.startswith("V_factor_")
            x = value
            assert isinstance(x, ea.Expr)
            c, t = self.gen_expr(x)
            assert t == "V_a"
            append(f"# {name}")
            append_cy(f"cdef cc.PyComplex {name} = {c}")
            append_py(f"{name} = {c}")
            append(f"factors_view[{idx}] = {name}")
        append(f"# set flops")
        append(f"total_flops = len(props) * 144 * 2 * 8 + len(cms) * 9 * 2 * 8 + len(factors) * 2 * 8")
        append(f"# return")
        append(f"return total_flops, (props, cms, factors,)")
        self.indent -= 4

    def cexpr_function_eval(self):
        append = self.append
        append_cy = self.append_cy
        append_py = self.append_py
        cexpr = self.cexpr
        append(f"@timer_flops")
        append(f"def cexpr_function_eval(positions_dict, props, cms, factors):")
        self.indent += 4
        append(f"# load AMA props with proper format")
        append(f"props = [ load_prop(p) for p in props ]")
        append(f"# join the AMA props")
        append(f"ama_props = ama_list(*props)")
        append(f"# apply eval to the factors and AMA props")
        append(f"ama_val = ama_apply1(lambda x_props: cexpr_function_eval_with_props(positions_dict, x_props, cms, factors), ama_props)")
        append(f"# set flops")
        append(f"total_flops = ama_counts(ama_val) * total_sloppy_flops")
        append(f"# return")
        append(f"return total_flops, ama_val")
        self.indent -= 4

    def cexpr_function_eval_with_props(self):
        append = self.append
        append_cy = self.append_cy
        append_py = self.append_py
        cexpr = self.cexpr
        append(f"@timer_flops")
        append_cy(f"@cython.boundscheck(False)")
        append_cy(f"@cython.wraparound(False)")
        append_cy(f"def cexpr_function_eval_with_props(dict positions_dict, list props, list cms, cc.PyComplex[:] factors_view):")
        append_py(f"def cexpr_function_eval_with_props(positions_dict, props, cms, factors_view):")
        self.indent += 4
        append(f"# set positions")
        for position_var in cexpr.positions:
            if position_var in aff.auto_fac_funcs_list:
                append(f"{position_var} = aff.{position_var}")
            else:
                append(f"{position_var} = positions_dict['{position_var}']")
        append(f"# set props")
        for idx, (name, value,) in enumerate(cexpr.variables_prop):
            append_cy(f"cdef cc.WilsonMatrix* p_{name} = &(<qu.WilsonMatrix>props[{idx}]).xx")
            append_py(f"p_{name} = props[{idx}]")
        append(f"# set cms")
        for idx, (name, value,) in enumerate(cexpr.variables_color_matrix):
            append_cy(f"cdef cc.ColorMatrix* p_{name} = &(<qu.ColorMatrix>cms[{idx}]).xx")
            append_py(f"p_{name} = cms[{idx}]")
        append(f"# set factors")
        for idx, (name, value,) in enumerate(cexpr.variables_factor):
            append_cy(f"cdef cc.PyComplex {name} = factors_view[{idx}]")
            append_py(f"{name} = factors_view[{idx}]")
        append(f"# compute products")
        for name, value in cexpr.variables_prod:
            assert name.startswith("V_prod_")
            x = value
            assert isinstance(x, tuple)
            assert len(x) == 2
            c, t = self.gen_expr_prod_list(x)
            assert t == get_var_name_type(name)
            if t == "V_G":
                append_cy(f"cdef cc.SpinMatrix {name} = {c}")
                append_py(f"{name} = {c}")
            elif t == "V_S":
                append_cy(f"cdef cc.WilsonMatrix {name} = {c}")
                append_py(f"{name} = {c}")
            else:
                assert False
        append(f"# compute traces")
        for name, value in cexpr.variables_tr:
            assert name.startswith("V_tr_")
            x = value
            assert isinstance(x, Op)
            assert x.otype == "Tr"
            c, t = self.gen_expr(x)
            assert t == "V_a"
            append_cy(f"cdef cc.PyComplex {name} = cc.py_complex_cast({c})")
            append_py(f"{name} = {c}")
        append(f"# set terms")
        for name, term in cexpr.named_terms:
            x = term
            if x.coef == 1:
                c_ops = x.c_ops
            else:
                c_ops = [ x.coef, ] + x.c_ops
            c, t = self.gen_expr_prod_list(c_ops)
            append_cy(f"cdef cc.PyComplex {name} = {c}")
            append_py(f"{name} = {c}")
        append(f"# declare exprs")
        append_cy(f"cdef numpy.ndarray[numpy.complex128_t] exprs")
        append(f"exprs = np.zeros({len(cexpr.named_exprs)}, dtype = np.complex128)")
        append_cy(f"cdef cc.PyComplex[:] exprs_view = exprs")
        append_py(f"exprs_view = exprs")
        append(f"# set exprs")
        def show_coef_term(coef, tname):
            coef, t = self.gen_expr(coef)
            assert t == "V_a"
            if coef == "1":
                return f"{tname}"
            else:
                return f"({coef}) * {tname}"
        append_cy(f"cdef cc.PyComplex expr")
        for idx, (name, expr,) in enumerate(cexpr.named_exprs):
            name = name.replace("\n", "  ")
            append(f"# {idx} name='{name}' ")
            append(f"expr = 0")
            for coef, tname in expr:
                s = show_coef_term(coef, tname)
                append(f"expr += {s}")
            append(f"exprs_view[{idx}] = expr")
        append(f"# set flops")
        append(f"total_flops = total_sloppy_flops")
        append(f"# return")
        append(f"return total_flops, exprs")
        self.indent -= 4

    def total_flops(self):
        append = self.append
        append(f"# Total flops per sloppy call is: {self.total_sloppy_flops}")
        append(f"total_sloppy_flops = {self.total_sloppy_flops}")

#### ----

if __name__ == "__main__":
    expr = Qb("d", "x1", "s1", "c1") * G(5, "s1", "s2") * Qv("u", "x1", "s2", "c1") * Qb("u", "x2", "s3", "c2") * G(5, "s3", "s4") * Qv("d", "x2", "s4", "c2")
    print(expr)
    print()
    expr = simplified(contract_expr(expr))
    print(expr)
    print()
    cexpr = copy.deepcopy(mk_cexpr(expr))
    print(cexpr)
    print()
    cexpr.optimize()
    print(cexpr)
    print()
    print(display_cexpr(cexpr))
    print()
    print(expr)
    print()
    cexpr = contract_simplify_compile(expr, is_isospin_symmetric_limit = True)
    print(display_cexpr(cexpr))
    print()
    cexpr.optimize()
    print(display_cexpr(cexpr))
    print(cexpr_code_gen_py(cexpr))



import qlat as q
import gc
import os
import pprint

from . import rbc_ukqcd_params as rup
from .rbc_ukqcd_params import get_total_site
from .rbc_ukqcd_params import load_config, load_config_lazy

def get_fermion_params(job_tag, inv_type, inv_acc):
    return rup.dict_params[job_tag]["fermion_params"][inv_type][inv_acc]

def get_ls_from_fermion_params(fermion_params):
    if "omega" in fermion_params:
        return len(fermion_params["omega"])
    else:
        return fermion_params["Ls"]

def get_lanc_params(job_tag, inv_type, inv_acc = 0):
    return rup.dict_params[job_tag]["lanc_params"][inv_type][inv_acc]

def get_clanc_params(job_tag, inv_type, inv_acc = 0):
    return rup.dict_params[job_tag]["clanc_params"][inv_type][inv_acc]

@q.timer_verbose
def mk_eig(gf, job_tag, inv_type, inv_acc = 0):
    import qlat_gpt as qg
    import gpt as g
    qtimer = q.Timer(f"py:mk_eig({job_tag},{inv_type},{inv_acc})", True)
    qtimer.start()
    gpt_gf = g.convert(qg.gpt_from_qlat(gf), g.single)
    parity = g.odd
    params = get_lanc_params(job_tag, inv_type, inv_acc)
    q.displayln_info(f"mk_eig: job_tag={job_tag} inv_type={inv_type} inv_acc={inv_acc}")
    q.displayln_info(f"mk_eig: params=\n{pprint.pformat(params)}")
    fermion_params = params["fermion_params"]
    if "omega" in fermion_params:
        qm = g.qcd.fermion.zmobius(gpt_gf, fermion_params)
    else:
        qm = g.qcd.fermion.mobius(gpt_gf, fermion_params)
    w = g.qcd.fermion.preconditioner.eo2_ne(parity=parity)(qm)
    def make_src(rng):
        src = g.vspincolor(qm.F_grid_eo)
        # src[:] = g.vspincolor([[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]])
        rng.cnormal(src)
        src.checkerboard(parity)
        return src
    pit = g.algorithms.eigen.power_iteration(**params["pit_params"])
    pit_ev, _, _ = pit(w.Mpc, make_src(g.random("lanc")));
    q.displayln_info(f"mk_eig: pit_ev={pit_ev}")
    #
    cheby = g.algorithms.polynomial.chebyshev(params["cheby_params"])
    irl = g.algorithms.eigen.irl(params["irl_params"])
    evec, ev = irl(cheby(w.Mpc), make_src(g.random("lanc")))
    evals, eps2 = g.algorithms.eigen.evals(w.Mpc, evec, real=True)
    g.mem_report()
    #
    qtimer.stop()
    return evec, evals

@q.timer_verbose
def mk_ceig(gf, job_tag, inv_type, inv_acc = 0):
    import qlat_gpt as qg
    import gpt as g
    qtimer = q.Timer(f"py:mk_ceig({job_tag},{inv_type},{inv_acc})", True)
    qtimer.start()
    gpt_gf = g.convert(qg.gpt_from_qlat(gf), g.single)
    parity = g.odd
    params = get_lanc_params(job_tag, inv_type, inv_acc)
    cparams = get_clanc_params(job_tag, inv_type, inv_acc)
    assert cparams["nbasis"] <= params["irl_params"]["Nstop"]
    fermion_params = params["fermion_params"]
    q.displayln_info(f"mk_ceig: job_tag={job_tag} inv_type={inv_type} inv_acc={inv_acc}")
    q.displayln_info(f"mk_ceig: params=\n{pprint.pformat(params)}")
    if "omega" in fermion_params:
        qm = g.qcd.fermion.zmobius(gpt_gf, fermion_params)
    else:
        qm = g.qcd.fermion.mobius(gpt_gf, fermion_params)
    w = g.qcd.fermion.preconditioner.eo2_ne(parity=parity)(qm)
    def make_src(rng):
        src = g.vspincolor(qm.F_grid_eo)
        # src[:] = g.vspincolor([[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]])
        rng.cnormal(src)
        src.checkerboard(parity)
        return src
    pit = g.algorithms.eigen.power_iteration(**params["pit_params"])
    pit_ev, _, _ = pit(w.Mpc, make_src(g.random("lanc")));
    q.displayln_info(f"mk_ceig: pit_ev={pit_ev}")
    #
    cheby = g.algorithms.polynomial.chebyshev(params["cheby_params"])
    irl = g.algorithms.eigen.irl(params["irl_params"])
    evec, ev = irl(cheby(w.Mpc), make_src(g.random("lanc")))
    evals, eps2 = g.algorithms.eigen.evals(w.Mpc, evec, real=True)
    g.mem_report()
    #
    inv = g.algorithms.inverter
    #
    q.displayln_info(f"mk_ceig: cparams=\n{pprint.pformat(cparams)}")
    #
    grid_coarse = g.block.grid(qm.F_grid_eo, [ get_ls_from_fermion_params(fermion_params) ] + cparams["block"])
    nbasis = cparams["nbasis"]
    basis = evec[0:nbasis]
    b = g.block.map(grid_coarse, basis)
    for i in range(2):
        b.orthonormalize()
    del evec
    gc.collect()
    #
    ccheby = g.algorithms.polynomial.chebyshev(cparams["cheby_params"])
    cop = b.coarse_operator(ccheby(w.Mpc))
    #
    cstart = g.vcomplex(grid_coarse, nbasis)
    cstart[:] = g.vcomplex([1] * nbasis, nbasis)
    eps2 = g.norm2(cop * cstart - b.project * ccheby(w.Mpc) * b.promote * cstart) / g.norm2(cstart)
    g.message(f"Test coarse-grid promote/project cycle: {eps2}")
    cirl = g.algorithms.eigen.irl(cparams["irl_params"])
    cevec, cev = cirl(cop, cstart)
    #
    smoother = inv.cg(cparams["smoother_params"])(w.Mpc)
    smoothed_evals = []
    tmpf = g.lattice(basis[0])
    for i, cv in enumerate(cevec):
        tmpf @= smoother * b.promote * cv
        evals = g.algorithms.eigen.evals(
            w.Mpc, [tmpf], calculate_eps2 = False, real=True
        )
        smoothed_evals = smoothed_evals + evals
    g.mem_report()
    #
    qtimer.stop()
    return basis, cevec, smoothed_evals

@q.timer_verbose
def get_smoothed_evals(basis, cevec, gf, job_tag, inv_type, inv_acc = 0):
    import qlat_gpt as qg
    import gpt as g
    gpt_gf = g.convert(qg.gpt_from_qlat(gf), g.single)
    parity = g.odd
    params = get_lanc_params(job_tag, inv_type, inv_acc)
    cparams = get_clanc_params(job_tag, inv_type, inv_acc)
    assert cparams["nbasis"] <= params["irl_params"]["Nstop"]
    fermion_params = params["fermion_params"]
    if "omega" in fermion_params:
        qm = g.qcd.fermion.zmobius(gpt_gf, fermion_params)
    else:
        qm = g.qcd.fermion.mobius(gpt_gf, fermion_params)
    w = g.qcd.fermion.preconditioner.eo2_ne(parity=parity)(qm)
    inv = g.algorithms.inverter
    grid_coarse = g.block.grid(qm.F_grid_eo, [ get_ls_from_fermion_params(fermion_params) ] + cparams["block"])
    b = g.block.map(grid_coarse, basis)
    smoother = inv.cg(cparams["smoother_params"])(w.Mpc)
    smoothed_evals = []
    tmpf = g.lattice(basis[0])
    for i, cv in enumerate(cevec):
        tmpf @= smoother * b.promote * cv
        evals = g.algorithms.eigen.evals(
            w.Mpc, [tmpf], calculate_eps2 = False, real=True
        )
        smoothed_evals = smoothed_evals + evals
    return smoothed_evals

@q.timer_verbose
def save_ceig(path, eig, job_tag, inv_type = 0, inv_acc = 0):
    import gpt as g
    if path is None:
        return
    save_params = get_clanc_params(job_tag, inv_type, inv_acc)["save_params"]
    nsingle = save_params["nsingle"]
    mpi = save_params["mpi"]
    fmt = g.format.cevec({"nsingle": nsingle, "mpi": [ 1 ] + mpi, "max_read_blocks": 8})
    q.mk_file_dirs_info(path)
    g.save(path, eig, fmt);

@q.timer_verbose
def load_eig_lazy(path, job_tag, inv_type = 0, inv_acc = 0):
    """
    return ``None'' or a function ``load_eig''
    ``load_eig()'' return the ``eig''
    """
    import qlat_gpt as qg
    import gpt as g
    if path is None:
        q.displayln_info(f"load_eig_lazy: path is '{path}'")
        return None
    if not q.does_file_exist_qar_sync_node(os.path.join(path, "metadata.txt")):
        q.displayln_info(f"load_eig_lazy: '{path}' has not metadata.")
        return None
    if not q.does_file_exist_qar_sync_node(os.path.join(path, "eigen-values.txt")):
        q.displayln_info(f"load_eig_lazy: '{path}' has not eigen-values.")
        return None
    if not q.does_file_exist_qar_sync_node(os.path.join(path, "00/0000000000.compressed")):
        q.displayln_info(f"load_eig_lazy: '{path}' has not data file '00/0000000000.compressed'.")
        return None
    #
    @q.timer_verbose
    def load_eig():
        total_site = get_total_site(job_tag)
        fermion_params = get_fermion_params(job_tag, inv_type, inv_acc)
        grids = qg.get_fgrid(total_site, fermion_params)
        eig = g.load(path, grids = grids)
        return eig
    #
    return q.lazy_call(load_eig)

def get_cg_mp_maxiter(job_tag, inv_type, inv_acc):
    cg_params = rup.dict_params[job_tag].get(f"cg_params-{inv_type}-{inv_acc}")
    if cg_params is not None:
        maxiter = cg_params.get("maxiter")
        if maxiter is not None:
            return maxiter
    if job_tag[:5] == "test-":
        maxiter = 50
    elif inv_acc == 2:
        maxiter = 500
    elif inv_type == 0:
        maxiter = 200
    elif inv_type in [ 1, 2, ]:
        maxiter = 300
    else:
        maxiter = 100
        raise Exception("get_cg_mp_maxiter")
    return maxiter

@q.timer_verbose
def mk_gpt_inverter(gf, job_tag, inv_type, inv_acc, *,
        gt = None,
        mpi_split = None,
        n_grouped = 1,
        eig = None,
        eps = 1e-8,
        qtimer = True):
    import qlat_gpt as qg
    import gpt as g
    if mpi_split is None:
        mpi_split = g.default.get_ivec("--mpi_split", None, 4)
        if mpi_split is not None:
            n_grouped = g.default.get_int("--grouped", 4)
    q.displayln_info(f"mk_gpt_inverter: job_tag={job_tag} inv_type={inv_type} inv_acc={inv_acc} mpi_split={mpi_split} n_grouped={n_grouped}")
    gpt_gf = qg.gpt_from_qlat(gf)
    pc = g.qcd.fermion.preconditioner
    if inv_type in [ 0, 1, 2, ]:
        params = get_fermion_params(job_tag, inv_type, inv_acc)
        if eig is not None:
            # may need madwf
            params0 = get_fermion_params(job_tag, inv_type, inv_acc = 0)
            is_madwf = get_ls_from_fermion_params(params) != get_ls_from_fermion_params(params0)
            if is_madwf and inv_type == 1:
                # do not use madwf for strange even when eig is available (do not use eig for exact solve)
                is_madwf = False
                eig = None
        else:
            is_madwf = False
        q.displayln_info(f"mk_gpt_inverter: set qm params={params}")
        if "omega" in params:
            qm = g.qcd.fermion.zmobius(gpt_gf, params)
        else:
            qm = g.qcd.fermion.mobius(gpt_gf, params)
        inv = g.algorithms.inverter
        cg_mp = inv.cg({"eps": eps, "maxiter": get_cg_mp_maxiter(job_tag, inv_type, inv_acc)})
        cg_pv_f = inv.cg({"eps": eps, "maxiter": 150})
        if mpi_split is None or mpi_split == False:
            cg_split = cg_mp
            n_grouped = 1
            mpi_split = None
        else:
            cg_split = inv.split(cg_mp, mpi_split = mpi_split)
            cg_pv_f = inv.split(cg_pv_f, mpi_split = mpi_split)
        q.displayln_info(f"mk_gpt_inverter: mpi_split={mpi_split} n_grouped={n_grouped}")
        if eig is not None:
            cg_defl = inv.coarse_deflate(eig[1], eig[0], eig[2])
            cg = inv.sequence(cg_defl, cg_split)
        else:
            cg = cg_split
        if "omega" in params and eig is None:
            pc_ne = pc.eo2_kappa_ne()
        else:
            pc_ne = pc.eo2_ne()
        if inv_type == 0:
            slv_5d = inv.preconditioned(pc_ne, cg)
        elif inv_type in [ 1, 2, ]:
            slv_5d = inv.preconditioned(pc_ne, cg)
        else:
            raise Exception("mk_gpt_inverter")
        q.displayln_info(f"mk_gpt_inverter: deal with is_madwf={is_madwf}")
        if is_madwf:
            gpt_gf_f = g.convert(gpt_gf, g.single)
            if "omega" in params0:
                qm0 = g.qcd.fermion.zmobius(gpt_gf_f, params0)
            else:
                qm0 = g.qcd.fermion.mobius(gpt_gf_f, params0)
            slv_5d_pv_f = inv.preconditioned(pc.eo2_ne(), cg_pv_f)
            slv_5d = pc.mixed_dwf(slv_5d, slv_5d_pv_f, qm0)
        if inv_acc == 0:
            maxiter = 1
        elif inv_acc == 1:
            maxiter = 2
        elif inv_acc == 2:
            maxiter = 200
        else:
            raise Exception("mk_gpt_inverter")
        q.sync_node()
        q.displayln_info(f"mk_gpt_inverter: eps={eps} max_cycle={maxiter}")
        slv_qm = qm.propagator(
                inv.defect_correcting(
                    inv.mixed_precision(
                        slv_5d, g.single, g.double),
                    eps=eps, maxiter=maxiter)).grouped(n_grouped)
        q.displayln_info(f"mk_gpt_inverter: make inv_qm")
        if qtimer is True:
            qtimer = q.Timer(f"py:inv({job_tag},{inv_type},{inv_acc})", True)
        elif qtimer is False:
            qtimer = q.TimerNone()
        inv_qm = qg.InverterGPT(inverter = slv_qm, qtimer = qtimer)
    else:
        raise Exception("mk_gpt_inverter")
    if gt is None:
        return inv_qm
    else:
        return q.InverterGaugeTransform(inverter = inv_qm, gt = gt)

@q.timer_verbose
def mk_qlat_inverter(gf, job_tag, inv_type, inv_acc, *, gt = None):
    qtimer = q.Timer(f"py:qinv({job_tag},{inv_type},{inv_acc})", True)
    if job_tag in ["24D", "32D"]:
        if inv_type == 0:
            fa = q.FermionAction(mass = 0.00107, m5 = 1.8, ls = 24, mobius_scale = 4.0)
            inv = q.InverterDomainWall(gf = gf, fa = fa, qtimer = qtimer)
            inv.set_stop_rsd(1e-8)
            inv.set_max_num_iter(200)
            if inv_acc == 0:
                maxiter = 1
            elif inv_acc == 1:
                maxiter = 2
            elif inv_acc == 2:
                maxiter = 50
            else:
                raise Exception("mk_qlat_inverter")
            inv.set_max_mixed_precision_cycle(maxiter)
        elif inv_type == 1:
            fa = q.FermionAction(mass = 0.0850, m5 = 1.8, ls = 24, mobius_scale = 4.0)
            inv = q.InverterDomainWall(gf = gf, fa = fa, qtimer = qtimer)
            inv.set_stop_rsd(1e-8)
            inv.set_max_num_iter(300)
            if inv_acc == 0:
                maxiter = 1
            elif inv_acc == 1:
                maxiter = 2
            elif inv_acc == 2:
                maxiter = 50
            else:
                raise Exception("mk_qlat_inverter")
            inv.set_max_mixed_precision_cycle(maxiter)
        else:
            raise Exception("mk_qlat_inverter")
    else:
        raise Exception("mk_qlat_inverter")
    if gt is None:
        return inv
    else:
        return q.InverterGaugeTransform(inverter = inv, gt = gt)

def mk_inverter(*args, **kwargs):
    return mk_gpt_inverter(*args, **kwargs)

@q.timer
def get_inv(gf, job_tag, inv_type, inv_acc, *, gt = None, mpi_split = None, n_grouped = 1, eig = None, eps = 1e-8, qtimer = True):
    tag = f"rbc_ukqcd.get_inv gf={id(gf)} {job_tag} inv_type={inv_type} inv_acc={inv_acc} gt={id(gt)} mpi_split={mpi_split} n_grouped={n_grouped} eig={id(eig)} eps={eps} qtimer={id(qtimer)}"
    if tag in q.cache_inv:
        return q.cache_inv[tag]["inv"]
    inv = mk_inverter(gf, job_tag, inv_type, inv_acc,
            gt = gt,
            mpi_split = mpi_split,
            n_grouped = n_grouped,
            eig = eig,
            eps = eps,
            qtimer = qtimer)
    q.cache_inv[tag] = {
            "inv": inv,
            "gf": gf,
            "gt": gt,
            "eig": eig,
            "qtimer": qtimer,
            }
    return inv



from .. import rbc_ukqcd_params as rup

from . import dict_params

import qlat as q

def get_param(*keys, default = None, dict_params = None):
    if dict_params is None:
        dict_params = rup.dict_params
    d = dict_params
    for key in keys:
        if key in d:
            d = d[key]
        else:
            return default
    return d

def set_param(*keys, value, dict_params = None):
    if dict_params is None:
        dict_params = rup.dict_params
    assert len(keys) >= 1
    d = dict_params
    for key in keys[:-1]:
        if key not in d:
            d[key] = dict()
        d = d[key]
    d[keys[-1]] = value

def get_total_site(job_tag : str):
    return get_param(job_tag, "total_site")

@q.timer_verbose
def mk_sample_gauge_field(job_tag, fn):
    rs = q.RngState(f"seed {job_tag} {fn}").split("mk_sample_gauge_field")
    total_site = get_total_site(job_tag)
    geo = q.Geometry(total_site, 1)
    gf = q.GaugeField(geo)
    gf.set_rand(rs, sigma = 0.25, n_step = 4)
    for i in range(4):
        q.gf_wilson_flow_step(gf, 0.05)
    gf.unitarize()
    return gf

@q.timer_verbose
def mk_sample_gauge_field_v2(total_site, tag):
    rs = q.RngState(f"seed {total_site} {tag}").split("mk_sample_gauge_field")
    geo = q.Geometry(total_site, 1)
    gf = q.GaugeField(geo)
    gf.set_rand(rs.split("gf.set_rand"), sigma = 0.25, n_step = 16)
    gf.unitarize()
    beta = 6.0
    ga = q.GaugeAction(beta)
    for traj in range(500):
        q.run_hmc_pure_gauge(gf, ga, traj, rs.split("run_hmc_pure_gauge"),
                             n_step = 6, is_always_accept = True)
    gf.unitarize()
    return gf

@q.timer_verbose
def mk_sample_gauge_field_v3(job_tag, fn):
    """
    depends on
    total_site = get_param(job_tag, "total_site")
    rand_n_step = get_param(job_tag, "mk_sample_gauge_field", "rand_n_step", default = 16)
    rand_sigma = get_param(job_tag, "mk_sample_gauge_field", "rand_sigma", default = 0.25)
    flow_n_step = get_param(job_tag, "mk_sample_gauge_field", "flow_n_step", default = 4)
    flow_size = get_param(job_tag, "mk_sample_gauge_field", "flow_size", default = 0.05)
    hmc_n_traj = get_param(job_tag, "mk_sample_gauge_field", "hmc_n_traj", default = 500)
    hmc_beta = get_param(job_tag, "mk_sample_gauge_field", "hmc_beta", default = 6.0)
    hmc_n_step = get_param(job_tag, "mk_sample_gauge_field", "hmc_n_step", default = 6)
    hmc_is_always_accept = get_param(job_tag, "mk_sample_gauge_field", "hmc_is_always_accept", default = True)
    """
    rs = q.RngState(f"seed {job_tag} {fn}").split("mk_sample_gauge_field")
    total_site = get_param(job_tag, "total_site")
    rand_n_step = get_param(job_tag, "mk_sample_gauge_field", "rand_n_step", default = 16)
    rand_sigma = get_param(job_tag, "mk_sample_gauge_field", "rand_sigma", default = 0.25)
    flow_n_step = get_param(job_tag, "mk_sample_gauge_field", "flow_n_step", default = 4)
    flow_size = get_param(job_tag, "mk_sample_gauge_field", "flow_size", default = 0.05)
    hmc_n_traj = get_param(job_tag, "mk_sample_gauge_field", "hmc_n_traj", default = 500)
    hmc_beta = get_param(job_tag, "mk_sample_gauge_field", "hmc_beta", default = 6.0)
    hmc_n_step = get_param(job_tag, "mk_sample_gauge_field", "hmc_n_step", default = 6)
    hmc_is_always_accept = get_param(job_tag, "mk_sample_gauge_field", "hmc_is_always_accept", default = True)
    geo = q.Geometry(total_site, 1)
    gf = q.GaugeField(geo)
    gf.set_rand(rs.split("gf.set_rand"), sigma = rand_sigma, n_step = rand_n_step)
    gf.unitarize()
    for i in range(flow_n_step):
        q.gf_wilson_flow_step(gf, flow_size)
    gf.unitarize()
    ga = q.GaugeAction(hmc_beta)
    for traj in range(hmc_n_traj):
        q.run_hmc_pure_gauge(gf, ga, traj, rs.split("run_hmc_pure_gauge"),
                             n_step = hmc_n_step,
                             is_always_accept = hmc_is_always_accept)
    gf.unitarize()
    return gf

@q.timer_verbose
def load_config(job_tag : str, fn : str):
    if not q.does_file_exist_qar_sync_node(fn):
        raise Exception(f"load_config '{fn}' does not exist.")
    gf = q.GaugeField()
    gf.load(fn)
    # gf = qg.load_gauge_field(fn)
    if (job_tag in dict_params) and ("load_config_params" in dict_params[job_tag]):
        params = dict_params[job_tag]["load_config_params"]
        twist_boundary_at_boundary = params["twist_boundary_at_boundary"]
        for mu in range(4):
            lmom = twist_boundary_at_boundary[mu]
            if lmom != 0.0:
                q.displayln_info(f"load_config fn='{fn}' twist_boundary_at_boundary lmom={lmom} mu={mu}")
                gf.twist_boundary_at_boundary(lmom, mu)
    gf.show_info()
    return gf

@q.timer_verbose
def load_config_lazy(job_tag : str, fn : str):
    if not q.does_file_exist_qar_sync_node(fn):
        return None
    return q.lazy_call(load_config, job_tag, fn)



# Author: Luchang Jin 2023

import qlat as q
import sys

if len(sys.argv) < 2:
    q.displayln_info("Usage: fields-properly-truncate [--check-all] [--only-check] path1 path2 ...")
    exit()

q.begin_with_mpi()

args = sys.argv[1:]

is_check_all = False
is_only_check = False

while True:
    if args[0] == "--check-all":
        is_check_all = True
        args = args[1:]
    elif args[0] == "--only-check":
        is_only_check = True
        args = args[1:]
    else:
        break

for path in args:
    tags = q.properly_truncate_fields(
            path, is_check_all = is_check_all, is_only_check = is_only_check)
    for tag in tags:
        q.displayln_info(tag)

q.timer_display()

q.end_with_mpi()

exit()



from qlat_utils import *
import qlat.c as c
from qlat.c import get_size_node, get_coor_node
from qlat.c import sync_node
from qlat.c import begin, end
from math import prod

import numpy as np

default_size_node_list = list(map(Coordinate, [
    [ 1, 1, 1, 1, ],
    [ 1, 1, 1, 2, ],
    [ 1, 1, 2, 2, ],
    [ 1, 2, 2, 2, ],
    [ 1, 2, 2, 2, ],
    [ 2, 2, 2, 2, ],
    [ 2, 2, 2, 4, ],
    [ 2, 2, 4, 4, ],
    [ 2, 4, 4, 4, ],
    [ 4, 4, 4, 4, ],
    [ 4, 4, 4, 8, ],
    [ 4, 4, 8, 8, ],
    [ 4, 8, 8, 8, ],
    [ 8, 8, 8, 8, ],
    [ 8, 8, 8, 16, ],
    [ 1, 1, 1, 3, ],
    [ 1, 1, 2, 3, ],
    [ 1, 2, 2, 3, ],
    [ 2, 2, 2, 3, ],
    [ 2, 2, 2, 6, ],
    [ 2, 2, 4, 6, ],
    [ 2, 4, 4, 6, ],
    [ 4, 4, 4, 6, ],
    [ 4, 4, 4, 12, ],
    [ 4, 4, 8, 12, ],
    [ 4, 8, 8, 12, ],
    [ 8, 8, 8, 12, ],
    ]))

def begin_with_mpi(size_node_list = None):
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    num_node = comm.size
    id_node = comm.rank
    if size_node_list is None:
        size_node_list = []
    else:
        assert isinstance(size_node_list, list)
        size_node_list = list(map(Coordinate, size_node_list))
    size_node_list = size_node_list + default_size_node_list
    size_node = None
    for size_node_check in size_node_list:
        if prod(size_node_check.to_list()) == num_node:
            size_node = size_node_check
            break
    if size_node is None:
        if id_node == 0:
            displayln(size_node_list)
        comm.barrier()
        raise Exception("begin_with_mpi: size_node_list not match num_node")
    c.begin(id_node, size_node)

def end_with_mpi(is_preserving_cache = False):
    c.end(is_preserving_cache)
    from mpi4py import MPI
    MPI.Finalize()

@timer
def glb_sum_np(x):
    """
    x does NOT change
    """
    from qlat_utils.lat_io import LatData
    shape = x.shape
    dtype = x.dtype
    l = list(x.flatten())
    ld = LatData()
    if dtype == np.dtype('float64'):
        ld.from_list(l, is_complex = False)
    elif dtype == np.dtype('int64'):
        ld.from_list(list(map(float, l)), is_complex = False)
    elif dtype == np.dtype('complex128'):
        ld.from_list(l, is_complex = True)
    else:
        displayln(dtype)
        assert False
    ld.glb_sum_in_place()
    return np.array(ld.to_list(), dtype = dtype).reshape(shape)

@timer
def glb_sum(x):
    """
    x does NOT change
    """
    if isinstance(x, float):
        return c.glb_sum_double(x)
    elif isinstance(x, complex):
        return c.glb_sum_complex(x)
    elif isinstance(x, int):
        return c.glb_sum_long(x)
    elif isinstance(x, np.ndarray):
        return glb_sum_np(x)
    elif isinstance(x, list):
        return [ glb_sum(x_i) for x_i in x ]
    elif isinstance(x, tuple):
        return tuple([ glb_sum(x_i) for x_i in x ])
    else:
        # possible types: q.LatData
        return x.glb_sum()

@timer_verbose
def show_machine():
    displayln(f"id_node: {get_id_node():4} / {get_num_node()}"
            f" ; coor_node: {str(get_coor_node()):9}"
            f" / {str(get_size_node())}")



#    Qlattice (https://github.com/jinluchang/qlattice)
#
#    Copyright (C) 2021
#
#    Author: Luchang Jin (ljin.luchang@gmail.com)
#    Author: Masaaki Tomii
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import qlat_utils as q
import copy
import sympy

try:
    from . import expr_arithmetic as ea
    from .expr_arithmetic import mk_sym
except:
    import expr_arithmetic as ea
    from expr_arithmetic import mk_sym

def mk_fac(x):
    """
    make an Expr obj (can be sympy expression)
    """
    return mk_expr(ea.mk_fac(x))

class Op:

    """
    self.otype
    """

    def __init__(self, otype : str):
        self.otype = otype

    def is_commute(self) -> bool:
        return True

    def show(self, is_multiply = False):
        return repr(self)

    def __repr__(self) -> str:
        return f"Op({self.otype})"

    def __add__(self, other):
        return mk_expr(self) + other

    __radd__ = __add__

    def __mul__(self, other):
        return mk_expr(self) * other

    def __rmul__(self, other):
        return mk_expr(other) * self

    def __neg__(self):
        return mk_expr(-1) * mk_expr(self)

    def __pos__(self):
        return self

    def __sub__(self, other):
        return mk_expr(self) + mk_expr(-1) * other

    def __rsub__(self, other):
        return mk_expr(other) + mk_expr(-1) * self

    def sort(self) -> None:
        pass

    def isospin_symmetric_limit(self) -> None:
        pass

### ------

class Qfield(Op):

    # self.f
    # self.p
    # self.s
    # self.c

    def __init__(self, otype : str, flavor : str, position : str, spin : str, color : str):
        Op.__init__(self, otype)
        self.f = flavor
        self.p = position
        self.s = spin
        self.c = color

    def is_commute(self) -> bool:
        return False

    def __repr__(self) -> str:
        if self.s == "auto" and self.c == "auto":
            return f"{self.otype}({self.f!r},{self.p!r})"
        else:
            return f"{self.otype}({self.f!r},{self.p!r},{self.s!r},{self.c!r})"

    def list(self):
        return [self.otype, self.f, self.p, self.s, self.c]

    def __eq__(self, other) -> bool:
        return self.list() == other.list()

### ------

class Qv(Qfield):

    """
    act as d / d Hb
    """

    def __init__(self, f, p, s, c):
        # interface function
        Qfield.__init__(self, "Qv", f, p, s, c)

### ------

class Qb(Qfield):

    """
    act as d / d Hv
    """

    def __init__(self, f, p, s, c):
        # interface function
        Qfield.__init__(self, "Qb", f, p, s, c)

### ------

class Hv(Qfield):

    def __init__(self, f, p, s, c):
        Qfield.__init__(self, "Hv", f, p, s, c)

### ------

class Hb(Qfield):

    def __init__(self, f, p, s, c):
        Qfield.__init__(self, "Hb", f, p, s, c)

### ------

class SHv(Qfield):

    def __init__(self, f, p, s, c):
        Qfield.__init__(self, "SHv", f, p, s, c)

### ------

class HbS(Qfield):

    def __init__(self, f, p, s, c):
        Qfield.__init__(self, "HbS", f, p, s, c)

### ------

class S(Op):

    """
    propagator
    #
    self.f
    self.p1
    self.p2
    self.s1
    self.s2
    self.c1
    self.c2
    """

    def __init__(self, flavor : str, p1 : str, p2 : str, s1 : str = "auto", s2 : str = "auto", c1 : str = "auto", c2 : str = "auto"):
        Op.__init__(self, "S")
        self.f = flavor
        self.p1 = p1
        self.p2 = p2
        self.s1 = s1
        self.s2 = s2
        self.c1 = c1
        self.c2 = c2

    def __repr__(self) -> str:
        if self.s1 == "auto" and self.s2 == "auto" and self.c1 == "auto" and self.c2 == "auto":
            return f"{self.otype}({self.f!r},{self.p1!r},{self.p2!r})"
        else:
            return f"{self.otype}({self.f!r},{self.p1!r},{self.p2!r},{self.s1!r},{self.s2!r},{self.c1!r},{self.c2!r})"

    def list(self):
        return [self.otype, self.f, self.p1, self.p2, self.s1, self.s2, self.c1, self.c2]

    def __eq__(self, other) -> bool:
        return self.list() == other.list()

    def isospin_symmetric_limit(self) -> None:
        # can also use these fictitious quark field to remove some unwanted disconnected diagrams
        if self.f in [ "u", "d", "u'", "d'", "u''", "d''", "u'''", "d'''", ]:
            self.f = "l"
        elif self.f in [ "s", "s'", "s''", "s'''", ]:
            self.f = "s"
        elif self.f in [ "c", "c'", "c''", "c'''", ]:
            self.f = "c"

### ------

class G(Op):

    """
    spin matrix
    #
    self.tag
    self.s1
    self.s2
    #
    tag = 0, 1, 2, 3, 5 for gamma matrices
    """

    def __init__(self, tag, s1 : str = "auto", s2 : str = "auto"):
        Op.__init__(self, "G")
        self.tag = tag
        self.s1 = s1
        self.s2 = s2

    def __repr__(self) -> str:
        if self.s1 == "auto" and self.s2 == "auto":
            return f"{self.otype}({self.tag!r})"
        else:
            return f"{self.otype}({self.tag!r},{self.s1!r},{self.s2!r})"

    def list(self):
        return [self.otype, self.tag, self.s1, self.s2]

    def __eq__(self, other) -> bool:
        return self.list() == other.list()

### ------

class U(Op):

    """
    color matrix
    #
    self.tag
    self.p
    self.mu
    self.c1
    self.c2
    """

    def __init__(self, tag, p, mu, c1 : str = "auto", c2 : str = "auto"):
        Op.__init__(self, "U")
        self.tag = tag
        self.p = p
        self.mu = mu
        self.c1 = c1
        self.c2 = c2

    def __repr__(self) -> str:
        if self.c1 == "auto" and self.c2 == "auto":
            return f"{self.otype}({self.tag!r},{self.p!r},{self.mu!r})"
        else:
            return f"{self.otype}({self.tag!r},{self.p!r},{self.mu!r},{self.c1!r},{self.c2!r})"

    def list(self):
        return [self.otype, self.tag, self.p, self.mu, self.c1, self.c2]

    def __eq__(self, other) -> bool:
        return self.list() == other.list()

### ------

def copy_op_index_auto(op : Op):
    if op.otype not in [ "S", "G", "U", ]:
        return op
    op = copy.copy(op)
    if op.otype in [ "S", "G", ]:
        op.s1 = "auto"
        op.s2 = "auto"
    if op.otype in [ "S", "U", ]:
        op.c1 = "auto"
        op.c2 = "auto"
    return op

class Tr(Op):

    """
    a collection of ops taking the trace
    #
    self.ops
    self.tag
    """

    def __init__(self, ops : list, tag = None):
        Op.__init__(self, "Tr")
        if tag is not None:
            # do not perform check if tag is set
            self.ops = ops
            self.tag = tag
            return
        for op in ops:
            assert op.is_commute()
            assert op.otype in [ "S", "G", "U", ]
        s = None
        c = None
        for op in ops + ops:
            if not check_trace_sc(op, s, c):
                raise Exception(f"ops={ops} tag={tag}")
            update_trace_sc(op, s, c)
        if s is not None and c is not None:
            self.tag = "sc"
        elif s is not None:
            self.tag = "s"
        elif c is not None:
            self.tag = "c"
        else:
            self.tag = ""
        for i, op in enumerate(ops):
            ops[i] = copy_op_index_auto(op)
        self.ops = ops

    def __repr__(self) -> str:
        return f"{self.otype}({self.ops!r},{self.tag!r})"

    def list(self):
        return [self.otype, self.tag, self.ops]

    def __eq__(self, other) -> bool:
        return self.list() == other.list()

    def sort(self):
        ops = self.ops
        if len(ops) > 1:
            self.ops = sorted([ ops[i:] + ops[:i] for i in range(len(ops)) ], key = repr)[0]

    def isospin_symmetric_limit(self) -> None:
        for op in self.ops:
            op.isospin_symmetric_limit()

### ------

def pick_one(xs, i):
    return xs[i], xs[:i] + xs[i + 1:]

def check_trace_spin_index(ops : list, s : str):
    count1 = 0
    count2 = 0
    i1 = None
    i2 = None
    for i, op in enumerate(ops):
        if op.otype in [ "S", "G", ]:
            if op.s1 == s:
                i1 = i
                count1 += 1
            if op.s2 == s:
                i2 = i
                count2 += 1
    return count1 == 1 and count2 == 1, i1, i2

def check_trace_color_index(ops : list, c : str):
    count1 = 0
    count2 = 0
    i1 = None
    i2 = None
    for i, op in enumerate(ops):
        if op.otype in [ "S", "U", ]:
            if op.c1 == c:
                i1 = i
                count1 += 1
            if op.c2 == c:
                i2 = i
                count2 += 1
    return count1 == 1 and count2 == 1, i1, i2

def check_trace_op(ops : list, op : Op):
    if op.otype not in [ "S", "G", "U", ]:
        return False
    if op.otype in [ "S", "G", ]:
        if not check_trace_spin_index(ops, op.s1)[0]:
            return False
        if not check_trace_spin_index(ops, op.s2)[0]:
            return False
    if op.otype in [ "S", "U", ]:
        if not check_trace_color_index(ops, op.c1)[0]:
            return False
        if not check_trace_color_index(ops, op.c2)[0]:
            return False
    return True

def check_trace_sc(op, s, c):
    if op.otype in [ "S", "G", ]:
        if s is not None and s != op.s1:
            return False
    if op.otype in [ "S", "U", ]:
        if c is not None and c != op.c1:
            return False
    return True

def update_trace_sc(op, s, c):
    if op.otype in [ "S", "G", ]:
        s = op.s2
    if op.otype in [ "S", "U", ]:
        c = op.c2
    return s, c

def pick_trace_op(ops : list, masks : list, s, c):
    for i, op in enumerate(ops):
        if masks[i]:
            continue
        if op.otype in [ "S", "G", ]:
            if s is not None and s != op.s1:
                continue
        if op.otype in [ "S", "U", ]:
            if c is not None and c != op.c1:
                continue
        if not check_trace_op(ops, op):
            continue
        return i, op
    return None

def find_trace(ops : list):
    """
    return None or (Tr(tr_ops), remaining_ops,)
    """
    size = len(ops)
    for i, op in enumerate(ops):
        if not check_trace_op(ops, op):
            continue
        masks = [ False for op in ops ]
        tr_ops = []
        s = None
        c = None
        masks[i] = True
        tr_ops.append(op)
        s, c = update_trace_sc(op, s, c)
        while True:
            p_op = pick_trace_op(ops, masks, s, c)
            if p_op is None:
                for op2 in tr_ops:
                    if not check_trace_sc(op2, s, c):
                        break
                    update_trace_sc(op2, s, c)
                return Tr(tr_ops), [ op for i, op in enumerate(ops) if not masks[i] ]
            i2, op2 = p_op
            masks[i2] = True
            tr_ops.append(op2)
            s, c = update_trace_sc(op2, s, c)
    return None

def collect_traces(ops : list) -> list:
    trs = []
    while True:
        ft = find_trace(ops)
        if ft is None:
            return trs + ops
        tr, ops = ft
        trs.append(tr)

class Term:

    """
    self.coef
    self.c_ops
    self.a_ops
    """

    def __init__(self, c_ops, a_ops, coef = 1):
        self.coef = coef
        self.c_ops = c_ops
        self.a_ops = a_ops
        for op in c_ops:
            assert op.is_commute()
        for op in a_ops:
            assert not op.is_commute()

    def list(self):
        return [ self.coef, self.c_ops, self.a_ops, ]

    def __eq__(self, other) -> bool:
        return self.list() == other.list()

    def check_commute(self) -> bool:
        for op in self.c_ops:
            assert op.is_commute()
        for op in self.a_ops:
            assert not op.is_commute()

    def __imul__(self, factor):
        self.coef *= factor

    def __add__(self, other):
        return mk_expr(self) + other

    def __add__(self, other):
        return mk_expr(self) + other

    __radd__ = __add__

    def __mul__(self, other):
        return mk_expr(self) * other

    def __rmul__(self, other):
        return mk_expr(other) * self

    def __neg__(self):
        return mk_expr(-1) * mk_expr(self)

    def __pos__(self):
        return self

    def __sub__(self, other):
        return mk_expr(self) + mk_expr(-1) * other

    def __rsub__(self, other):
        return mk_expr(other) + mk_expr(-1) * self

    def show(self, is_multiply = False) -> str:
        return "*".join([ f"({self.coef})", ] + self.c_ops + self.a_ops)

    def __repr__(self) -> str:
        return f"Term({self.c_ops},{self.a_ops},{self.coef})"

    def sort(self) -> None:
        # only sort commutable factors
        for op in self.c_ops:
            op.sort()
        self.c_ops.sort(key = repr)

    def simplify_coef(self) -> None:
        self.coef = ea.coef_simplified(self.coef)

    def simplify_ea(self) -> None:
        self.coef = ea.simplified(self.coef)

    def collect_traces(self) -> None:
        self.c_ops = collect_traces(self.c_ops)

    def isospin_symmetric_limit(self) -> None:
        for op in self.c_ops:
            op.isospin_symmetric_limit()

### ------

def combine_two_terms(t1 : Term, t2 : Term, t1_sig : str, t2_sig : str):
    if t1_sig == t2_sig:
        coef = t1.coef + t2.coef
        if ea.is_zero(coef):
            return Term([], [], 0)
        else:
            return Term(t1.c_ops, t1.a_ops, coef)
    else:
        return None

class Expr:

    # self.description
    # self.terms

    def __init__(self, terms, description = None):
        self.description = description
        self.terms = terms

    def __imul__(self, factor):
        for term in self.terms:
            term *= factor
        description = f"{factor} * {self.show(True)}"

    def __add__(self, other):
        # if other is str, then it is used to set the description of the result expr
        if isinstance(other, str):
            return Expr(self.terms, other)
        # otherwise return self + other
        other = mk_expr(other)
        terms = self.terms + other.terms
        return Expr(terms, f"+{self.show()} + {other.show()}")

    __radd__ = __add__

    def __mul__(self, other):
        other = mk_expr(other)
        terms = []
        for t1 in self.terms:
            for t2 in other.terms:
                coef = t1.coef * t2.coef
                if coef == 0:
                    continue
                t = Term(t1.c_ops + t2.c_ops, t1.a_ops + t2.a_ops, coef)
                terms.append(t)
        return Expr(terms, f"{self.show(True)} * {other.show(True)}")

    def __rmul__(self, other):
        return mk_expr(other) * self

    def __neg__(self):
        return mk_expr(-1) * mk_expr(self)

    def __pos__(self):
        return self

    def __sub__(self, other):
        return mk_expr(self) + mk_expr(-1) * other

    def __rsub__(self, other):
        return mk_expr(other) + mk_expr(-1) * self

    def show(self, is_multiply = False):
        if self.description is not None:
            assert isinstance(self.description, str)
            if len(self.description) >= 1 and self.description[0] == "+":
                if is_multiply:
                    return f"( {self.description[1:]} )"
                else:
                    return f"{self.description[1:]}"
            else:
                return self.description
        else:
            return repr(self)

    def __repr__(self) -> str:
        if self.description is None:
            return f"Expr({self.terms})"
        else:
            return f"Expr({self.terms},{self.description!r})"

    @q.timer
    def sort(self) -> None:
        for term in self.terms:
            term.sort()
        self.terms.sort(key = repr)

    @q.timer
    def simplify_coef(self) -> None:
        for t in self.terms:
            t.simplify_coef()
        self.drop_zeros()

    @q.timer
    def simplify_ea(self) -> None:
        for t in self.terms:
            t.simplify_ea()
        self.drop_zeros()

    @q.timer
    def combine_terms(self) -> None:
        self.terms = combine_terms_expr(self).terms

    @q.timer
    def drop_zeros(self) -> None:
        self.terms = drop_zero_terms(self).terms

    def round(self, ndigit : int = 20) -> None:
        # interface function
        sexpr = copy.deepcopy(self)
        for term in sexpr.terms:
            coef = term.coef
            term.coef = term.coef.evalf(ndigit)
        sexpr.terms = drop_zero_terms(sexpr).terms
        return sexpr

    @q.timer
    def collect_traces(self) -> None:
        for term in self.terms:
            term.collect_traces()

    @q.timer
    def isospin_symmetric_limit(self) -> None:
        for term in self.terms:
            term.isospin_symmetric_limit()

    @q.timer
    def simplify(self, *, is_isospin_symmetric_limit : bool = True) -> None:
        # interface function
        if is_isospin_symmetric_limit:
            self.isospin_symmetric_limit()
        self.sort()
        self.combine_terms()
        self.collect_traces()
        self.sort()
        self.combine_terms()
        self.simplify_ea()

### ------

def simplified(expr : Expr, *, is_isospin_symmetric_limit : bool = True) -> Expr:
    # interface function
    sexpr = copy.deepcopy(expr)
    sexpr.simplify(is_isospin_symmetric_limit = is_isospin_symmetric_limit)
    return sexpr

def mk_expr(x) -> Expr:
    if isinstance(x, Op):
        if x.is_commute():
            return Expr([Term([x,], [], 1),], f"{x}")
        else:
            return Expr([Term([], [x,], 1),], f"{x}")
    elif isinstance(x, Term):
        return Expr([x,], f"x.show()")
    elif isinstance(x, Expr):
        return x
    elif isinstance(x, (int, float, complex, sympy.Basic, ea.Expr)):
        return Expr([Term([], [], x),], f"({x})")
    else:
        print(x)
        assert False

@q.timer
def combine_terms_expr(expr : Expr) -> Expr:
    if not expr.terms:
        return expr
    def get_sig(t):
        return f"{t.c_ops},{t.a_ops}"
    zero_term = Term([], [], 0)
    zero_term_sig = get_sig(zero_term)
    signatures = [ get_sig(t) for t in expr.terms ]
    terms = []
    term = expr.terms[0]
    term_sig = signatures[0]
    for t, t_sig in zip(expr.terms[1:], signatures[1:]):
        if ea.is_zero(term.coef):
            term = t
            term_sig = t_sig
        else:
            ct = combine_two_terms(t, term, t_sig, term_sig)
            if ct is None:
                terms.append(term)
                term = t
                term_sig = t_sig
            elif ea.is_zero(ct.coef):
                term = zero_term
                term_sig = zero_term_sig
            else:
                term = ct
    if not ea.is_zero(term.coef):
        terms.append(term)
    return Expr(terms, expr.description)

@q.timer
def drop_zero_terms(expr : Expr) -> Expr:
    terms = []
    for t in expr.terms:
        if not ea.is_zero(t.coef):
            terms.append(t)
    return Expr(terms, expr.description)

def op_derivative_exp(op : Op):
    if op.otype == "Qv":
        return Term([], [SHv(op.f, op.p, op.s, op.c),], -1)
    elif op.otype == "Qb":
        return Term([], [HbS(op.f, op.p, op.s, op.c),], 1)
    else:
        return None

def op_derivative_op(op : Op, op1 : Op):
    if op.otype == "Qv" and op1.otype == "HbS" and op.f == op1.f:
        return Term([S(op.f, op.p, op1.p, op.s, op1.s, op.c, op1.c),], [], 1)
    elif op.otype == "Qb" and op1.otype == "SHv" and op.f == op1.f:
        return Term([S(op.f, op1.p, op.p, op1.s, op.s, op1.c, op.c),], [], 1)
    else:
        return None

def flip_sign(i : int) -> int:
    if i % 2 == 0:
        return 1
    else:
        return -1

def op_derivative_term(op : Op, term : Term) -> Expr:
    coef = term.coef
    c_ops = term.c_ops
    a_ops = term.a_ops
    terms = []
    for i in range(len(a_ops)):
        op1 = a_ops[i]
        dop1 = op_derivative_op(op, op1)
        if dop1 is not None:
            sign = flip_sign(i)
            terms.append(Term(dop1.c_ops + c_ops, a_ops[:i] + dop1.a_ops + a_ops[i+1:], sign * dop1.coef * coef))
    de = op_derivative_exp(op)
    if de is not None:
        sign = flip_sign(len(a_ops))
        terms.append(Term(de.c_ops + c_ops, a_ops + de.a_ops, sign * de.coef * coef))
    return Expr(terms)

def op_push_term(op : Op, term : Term) -> Expr:
    if op.otype == "Qv" or op.otype == "Qb":
        return op_derivative_term(op, term)
    else:
        coef = term.coef
        c_ops = term.c_ops
        a_ops = term.a_ops
        if op.is_commute():
            return Expr([Term([op,] + c_ops, a_ops, coef),])
        else:
            return Expr([Term(c_ops, [op,] + a_ops, coef),])

def op_push_expr(op : Op, expr : Expr) -> Expr:
    terms = []
    for term in expr.terms:
        terms += op_push_term(op, term).terms
    return Expr(terms)

def is_hop(op : Op) -> bool:
    if op.otype == "SHv" or op.otype == "HbS":
        return True
    if op.otype == "Hv" or op.otype == "Hb":
        return True
    return False

def has_hops(term : Term, count_limit : int = 0) -> bool:
    c = 0
    for op in term.a_ops:
        if is_hop(op):
            c += 1
            if c > count_limit:
                return True
    return False

def remove_hops(expr : Expr, count_limit : int = 0) -> Expr:
    terms = []
    for term in expr.terms:
        if not has_hops(term, count_limit):
            terms.append(term)
    return Expr(terms)

def contract_term(term : Term) -> Expr:
    coef = term.coef
    c_ops = term.c_ops
    a_ops = term.a_ops
    n_a_ops = len(a_ops)
    expr = Expr([Term(c_ops, [], coef),])
    for idx, op in enumerate(reversed(a_ops)):
        expr = op_push_expr(op, expr)
        expr = remove_hops(expr, n_a_ops - idx - 1)
    return expr

@q.timer
def contract_expr(expr: Expr) -> Expr:
    """
    interface function
    does not change expr
    """
    all_terms = []
    for term in expr.terms:
        all_terms += contract_term(term).terms
    return Expr(all_terms, f"< {expr.show()} >")

### ------

def S_l(p1, p2):
    return mk_expr(S('l', p1, p2)) + f"S_l({p1},{p2})"

def S_s(p1, p2):
    return mk_expr(S('s', p1, p2)) + f"S_s({p1},{p2})"

def S_c(p1, p2):
    return mk_expr(S('c', p1, p2)) + f"S_c({p1},{p2})"

def tr(expr):
    if isinstance(expr, Term):
        term = expr
        assert term.a_ops == []
        return Term([ Tr(term.c_ops), ], [], term.coef)
    elif isinstance(expr, Expr):
        return sum(map(tr, expr.terms)) + f"tr( {expr.show()} )"
    else:
        assert False

gamma_x = mk_expr(G(0)) + f"gamma_x"

gamma_y = mk_expr(G(1)) + f"gamma_y"

gamma_z = mk_expr(G(2)) + f"gamma_z"

gamma_t = mk_expr(G(3)) + f"gamma_t"

gamma_5 = mk_expr(G(5)) + f"gamma_5"

def gamma(tag):
    return mk_expr(G(tag)) + f"gamma({tag})"

def gamma_va(tag):
    assert isinstance(tag, int)
    if tag in [ 0, 1, 2, 3, ]:
        return mk_expr(G(tag)) + f"gamma({tag})"
    elif tag in [ 4, 5, 6, 7, ]:
        return mk_expr(G(tag - 4)) * mk_expr(G(5)) + f"gamma({tag-4})*gamma_5"
    else:
        assert False

### ------

if __name__ == "__main__":
    expr = (1
            * Qb("d", "x1", "s1", "c1")
            * G(5, "s1", "s2")
            * Qv("u", "x1", "s2", "c1")
            * Qb("u", "x2", "s3", "c2")
            * G(5, "s3", "s4")
            * Qv("d", "x2", "s4", "c2"))
    print(expr)
    c_expr = contract_expr(expr)
    c_expr.simplify(is_isospin_symmetric_limit = False)
    print(c_expr)
    c_expr_check = Expr([Term([Tr([G('5'), S('d','x2','x1'), G('5'), S('u','x1','x2')],'sc')],[],(-1+0j))]) - c_expr
    c_expr_check.simplify(is_isospin_symmetric_limit = False)
    print(c_expr_check)
    c_expr.simplify(is_isospin_symmetric_limit = True)
    print(c_expr)
    print(Qb("u", "x2", "s23", "c23") * Qv("u", "x1", "s11", "c11") - Qb("d", "x2", "s23", "c23") * Qv("d", "x1", "s11", "c11"))



import qlat.c as c

class FermionAction:

    def __init__(self, *, mass, ls, m5, mobius_scale = 1.0, omega = None):
        assert isinstance(mass, float)
        assert isinstance(ls, int)
        assert isinstance(m5, float)
        if omega is None:
            self.cdata = c.mk_fermion_action_mobius(mass, ls, m5, mobius_scale)
        else:
            assert isinstance(omega, list)
            assert ls == len(omega)
            self.cdata = c.mk_fermion_action_zmobius(mass, m5, omega)

    def __del__(self):
        assert isinstance(self.cdata, int)
        c.free_fermion_action(self)

    def __imatmul__(self, v1):
        assert isinstance(v1, FermionAction)
        c.set_fermion_action(self, v1)
        return self

    def mass(self):
        return c.get_mass_fermion_action(self)

    def ls(self):
        return c.get_ls_fermion_action(self)

    def m5(self):
        return c.get_m5_fermion_action(self)

    def omega(self):
        return c.get_omega_fermion_action(self)

    def mobius_scale(self):
        return c.get_mobius_scale_fermion_action(self)
